# Run from the root project (the 'test' parent folder): python3 -m unittest test/test_importer.py
# or from the IDE
import os
import unittest
import logging
import sys
from pathlib import Path

from _pytest.mark.expression import TokenType

import kernpy as kp

logger = logging.getLogger()
logger.level = logging.INFO  # change it DEBUG to trace errors
logger.addHandler(logging.StreamHandler(sys.stdout))


class ImporterTestCase(unittest.TestCase):
    def doJustImportTest(self, kern_file) -> kp.Document:
        logging.info(f'Importing {kern_file}')
        importer = kp.Importer()
        document = importer.import_file(kern_file)
        return document

    def checkEquals(self, kern_file, expected_ekern, from_measure, to_measure) -> kp.Document:
        importer = kp.Importer()
        document = importer.import_file(kern_file)
        if len(importer.errors) > 0:
            logging.error(f'Import errors for {kern_file}')
            for error in importer.errors:
                logging.error(f'\t{error}')

        dot_exporter = kp.GraphvizExporter()
        filename = os.path.basename(kern_file)
        #TODO No sé por qué me da error de recursión
        #dot_exporter.export_to_dot(document.tree, f'/tmp/{filename}.dot')

        # Read the content of both files
        with open(expected_ekern, 'r') as file1:
            expected_content = file1.read()

        exported_ekern = kp.dumps(document,
                                  spine_types=['**kern'],
                                  from_measure=from_measure,
                                  to_measure=to_measure,
                                  include=kp.BEKERN_CATEGORIES,
                                  encoding=kp.Encoding.eKern)

        if exported_ekern != expected_content:
            logging.info('---- Expected content ----')
            logging.info('--------------------------')
            logging.info(expected_content)

            logging.info('---- Exported content ----')
            logging.info('--------------------------')
            logging.info(exported_ekern)

        self.assertEqual(expected_content, exported_ekern)
        return document

    def doEKernTest(self, kern_file, expected_measure_start_rows):
        """
        :param kern_file:
        :param expected_measure_start_rows: Rows after removing empty lines and line comments
        :return:
        """
        logging.info(f'Importing {kern_file} and checking the ekern')
        ekern = os.path.splitext(kern_file)[0] + '.ekrn'
        document = self.checkEquals(kern_file, ekern, None, None)

        measure_start_tree_stages = document.measure_start_tree_stages
        self.assertEqual(expected_measure_start_rows, measure_start_tree_stages)

    def doEKernMeasureToMeasureTest(self, kern_file, from_measure, to_measure):
        logging.info(f'Importing {kern_file} and checking the ekern')
        ekern = f'{os.path.splitext(kern_file)[0]}-m{from_measure}-to-m{to_measure}.ekrn'
        self.checkEquals(kern_file, ekern, from_measure, to_measure)

    # it loads a simple file
    def testReadMinimalKern(self):
        document = self.doJustImportTest('resource_dir/unit/minimal.krn')
        dot_exporter = kp.GraphvizExporter()
        dot_exporter.export_to_dot(document.tree, '/tmp/minimal.dot')
        # self.assertEqual(1, len(ts.files))

    def testClefs(self):
        self.doJustImportTest('resource_dir/unit/clefs.krn')
        # self.assertEqual(1, len(ts.files))

    def testOctaves(self):
        self.doJustImportTest('resource_dir/unit/octaves.krn')
        # self.assertEqual(1, len(ts.files))

    def testBars(self):
        self.doJustImportTest('resource_dir/unit/bars.krn')
        # self.assertEqual(1, len(ts.files))

    def testTime(self):
        self.doJustImportTest('resource_dir/unit/time.krn')
        # self.assertEqual(1, len(ts.files))

    def testMensurations(self):
        self.doJustImportTest('resource_dir/unit/mensurations.krn')
        # self.assertEqual(1, len(ts.files))

    def testAccidentals(self):
        self.doJustImportTest('resource_dir/unit/accidentals.krn')
        # self.assertEqual(1, len(ts.files))

    def testAccidentalsWithAlterationDisplay(self):
        self.doJustImportTest('resource_dir/unit/accidentals_alteration_display.krn')
        # self.assertEqual(1, len(ts.files))

    def testKey(self):
        self.doJustImportTest('resource_dir/unit/key.krn')
        # self.assertEqual(1, len(ts.files))

    def testKeyDesignation(self):
        self.doJustImportTest('resource_dir/unit/key_designation.krn')
        # self.assertEqual(1, len(ts.files))

    def testModal(self):
        self.doJustImportTest('resource_dir/unit/modal.krn')
        # self.assertEqual(1, len(ts.files))

    def testChords(self):
        self.doJustImportTest('resource_dir/unit/chords.krn')
        # self.assertEqual(1, len(ts.files))

    def testRhythm(self):
        self.doJustImportTest('resource_dir/unit/rhythm.krn')
        # self.assertEqual(1, len(ts.files))

    def testTies(self):
        self.doJustImportTest('resource_dir/unit/ties.krn')
        # self.assertEqual(1, len(ts.files))

    def testBeams(self):
        self.doJustImportTest('resource_dir/unit/beaming.krn')
        # self.assertEqual(1, len(ts.files))

    def testAutoBeam(self):
        self.doJustImportTest('resource_dir/unit/auto_beaming.krn')
        # self.assertEqual(1, len(ts.files))

    def testRests(self):
        self.doJustImportTest('resource_dir/unit/rests.krn')
        # self.assertEqual(1, len(ts.files))

    def testSlurs(self):
        self.doJustImportTest('resource_dir/unit/slurs.krn')
        # self.assertEqual(1, len(ts.files))

    def testArticulations(self):
        self.doJustImportTest('resource_dir/unit/articulations.krn')
        # self.assertEqual(1, len(ts.files))

    def testOrnaments(self):
        self.doJustImportTest('resource_dir/unit/ornaments.krn')
        # self.assertEqual(1, len(ts.files))

    def testHeader(self):
        importer = kp.Importer()
        document = importer.import_file('resource_dir/unit/headers.krn')
        self.assertEqual(8, document.get_spine_count())
        header_stage = document.tree.stages[document.header_stage]
        expected_importers = [kp.KernSpineImporter, kp.MensSpineImporter, kp.DynamSpineImporter, kp.DynSpineImporter,
                              kp.HarmSpineImporter, kp.RootSpineImporter, kp.TextSpineImporter, kp.FingSpineImporter]
        for node, expected_importer in zip(header_stage, expected_importers):
            importer = kp.createImporter(node.token.encoding)
            self.assertTrue(isinstance(importer, expected_importer))

    def doTestCountSpines(self, kern_file, expected_stage_count, expected_spine_counts):
        logging.info(f'Importing {kern_file}')
        importer = kp.Importer()
        document = importer.import_file(kern_file)
        dot_exporter = kp.GraphvizExporter()
        filename = os.path.basename(kern_file)
        dot_exporter.export_to_dot(document.tree, f'/tmp/{filename}.dot')
        self.assertEqual(expected_stage_count, len(document.tree.stages) - 1, "Stage count")  # -1 to remove the root
        self.assertEqual(document.get_spine_count(), len(expected_spine_counts), "Num. spines")

        spine_counts = []
        for node in document.get_header_stage():
            spine_counts.append(node.count_nodes_by_stage())

        for i in range(document.get_spine_count()):
            self.assertListEqual(expected_spine_counts[i], spine_counts[i], f"{kern_file}, spine #{i + 1}")

    def testSpines(self):
        # Tests extracted from the discussion in  https://github.com/humdrum-tools/vhv-documentation/issues/7#event-3236429526
        self.doTestCountSpines('resource_dir/spines/non_stacked_ends.krn', 12,
                               [[1, 1, 2, 2, 3, 3, 4, 4, 2, 2, 1, 1]])  # in this test, three spines merge into one
        self.doTestCountSpines('resource_dir/spines/non_stacked_ends_2.krn', 10,
                               [[1, 1, 1, 1, 1, 1, 2, 2, 1, 1], [1, 1, 2, 2, 3, 3, 6, 6, 1, 1],
                                [1, 1, 1, 1, 1, 1, 2, 2, 1,
                                 1]])  # the last join in VHV leads to just one main spine - we've added an issue

        self.doTestCountSpines('resource_dir/spines/1.krn', 18,
                               [[1, 1, 1, 1, 2, 2, 2, 1, 1, 1, 1, 2, 2, 2, 2, 2, 1, 1]])
        self.doTestCountSpines('resource_dir/spines/2.krn', 18, [[1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
                                                                 [1, 1, 1, 1, 2, 2, 2, 1, 1, 1, 1, 2, 2, 2, 2, 2, 1,
                                                                  1]])
        self.doTestCountSpines('resource_dir/spines/3.krn', 16, [[1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 1, 1],
                                                                 [1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 1, 1]])
        self.doTestCountSpines('resource_dir/spines/4.krn', 17, [[1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1],
                                                                 [1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1]])
        self.doTestCountSpines('resource_dir/spines/5.krn', 24,
                               [[1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1],
                                [1, 1, 1, 1, 2, 2, 2, 3, 3, 3, 1, 1, 2, 3, 3, 3, 3, 3, 3, 1, 1, 1, 1, 1]])

        self.doTestCountSpines('resource_dir/spines/spines-from-piano-joplin-bethena-start.krn', 23,
                               [[1, 1, 1, 1, 1, 2, 2, 1, 1, 2, 2, 1, 1, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1],
                                [1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 2, 2, 1, 1],
                                [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]])
        self.doTestCountSpines('resource_dir/spines/spines-piano-hummel-prelude67-15.krn', 19,
                               [[1, 1, 1, 1, 1, 1, 1, 1, 2, 3, 3, 3, 3, 1, 1, 2, 2, 1, 1],
                                [1, 1, 2, 2, 1, 1, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1],
                                [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]])

    def testLegacyTests(self):
        self.doEKernTest('resource_dir/legacy/base_tuplet.krn', [5])
        self.doEKernTest('resource_dir/legacy/guide02-example2-1.krn', [5, 8, 11, 15])
        self.doEKernTest('resource_dir/legacy/guide02-example2-3.krn', [8, 9, 18, 22, 30, 38])
        self.doEKernTest('resource_dir/legacy/guide02-example2-4.krn', [6, 12, 16, 23, 27, 33, 37, 47, 51])
        self.doEKernTest('resource_dir/legacy/guide06-example6-1.krn', [5, 18, 27])
        self.doEKernTest('resource_dir/legacy/guide06-example6-2.krn', [6, 15, 28, 41])
        self.doEKernTest('resource_dir/legacy/chor001.krn',
                         # rows with comments
                         [26, 27, 32, 37, 43, 46, 50, 55, 57, 60, 67, 74, 77, 82, 88, 93, 96, 102, 107, 114, 117, 122,
                          128, 130])
        # rows without comments
        #[11, 12, 17, 22, 28, 31, 35, 40, 42, 45, 52, 58, 61, 66, 72, 77, 80, 86, 91, 98, 101, 106, 112, 114])
        self.doJustImportTest(
            'resource_dir/legacy/chor009.krn')  # , [23, 32, 39, 48, 53, 57, 65, 74, 83, 90, 99, 107, 116, 122])
        self.doJustImportTest('resource_dir/legacy/chor048.krn')  # , [22, 27, 32, 41, 46, 56, 65, 74, 83, 91, 98])

    def testBoundingBoxes(self):
        self.doJustImportTest(
            'resource_dir/polish/test1/pl-wn--mus-iii-118-771--003_badarzewska-tekla--mazurka-brillante.krn')

    def testSamples(self):
        self.doJustImportTest('resource_dir/samples/bach-brandenburg-bwv1050a.krn')
        self.doJustImportTest('resource_dir/samples/bach-chorale-chor205.krn')
        self.doJustImportTest('resource_dir/samples/corelli-op01n12d.krn')
        self.doJustImportTest('resource_dir/samples/harmonized-song-erk052.krn')
        self.doJustImportTest('resource_dir/samples/haydn-quartet-op54n2-01.krn')
        self.doJustImportTest('resource_dir/samples/piano-beethoven-sonata21-3.krn')
        self.doJustImportTest('resource_dir/samples/piano-chopin-prelude28-17.krn')
        self.doJustImportTest('resource_dir/samples/piano-hummel-prelude67-15.krn')
        self.doJustImportTest('resource_dir/samples/piano-joplin-bethena.krn')
        self.doJustImportTest('resource_dir/samples/piano-mozart-sonata07-3.krn')
        self.doJustImportTest('resource_dir/samples/piano-scarlatti-L523K205.krn')
        self.doJustImportTest('resource_dir/samples/quartet-beethoven-quartet13-6.krn')
        self.doJustImportTest('resource_dir/samples/quartet-mozart-k590-04.krn')
        self.doJustImportTest('resource_dir/samples/unaccompanied-songs-nova073.krn')

    def testExtractMeasures(self):
        self.doEKernMeasureToMeasureTest(
            'resource_dir/grandstaff/5901766.krn', 24, 28)

        self.doEKernMeasureToMeasureTest('resource_dir/spines/2.krn', 2, 2)
        self.doEKernMeasureToMeasureTest('resource_dir/legacy/chor001.krn', 1, 3)

        self.doEKernMeasureToMeasureTest(
            'resource_dir/polish/test2/pl-wn--mus-iii-123-982--001-004_wieniawski-henryk--l-ecole-moderne-etudes-caprices-pour-violon-seul-op-10-4-le-staccato.krn',
            1, 2)
        self.doEKernMeasureToMeasureTest(
            'resource_dir/polish/test1/pl-wn--mus-iii-118-771--003_badarzewska-tekla--mazurka-brillante.krn', 1, 2)
        self.doEKernMeasureToMeasureTest(
            'resource_dir/polish/test1/pl-wn--mus-iii-118-771--003_badarzewska-tekla--mazurka-brillante.krn', 1, 3)
        self.doEKernMeasureToMeasureTest(
            'resource_dir/polish/test1/pl-wn--mus-iii-118-771--003_badarzewska-tekla--mazurka-brillante.krn', 1, 16)

        #self.doEKernMeasureToMeasureTest('resource_dir/polish/test2/pl-wn--mus-iii-123-982--001-004_wieniawski-henryk--l-ecole-moderne-etudes-caprices-pour-violon-seul-op-10-4-le-staccato.krn', 0, 1) # TODO: Correct it. It doesn't export all the required token_categories

    #TODO Joan: este test se puede hacer como el doEKernMeasureToMeasureTest, sin repetir tanto código...
    def test_extract_measures_same_measure(self):
        options = kp.ExportOptions(spine_types=['**kern'], kern_type=kp.Encoding.normalizedKern, from_measure=3,
                                   to_measure=3)
        importer = kp.Importer()
        document = importer.import_file('resource_dir/legacy/base_tuplet_longer.krn')

        exporter = kp.Exporter()
        real_ouputput = exporter.export_string(document, options)

        with open('resource_dir/legacy/base_tuplet_longer_m3-m3.krn', 'r') as f:
            expected_output = f.read()
        self.assertEqual(real_ouputput, expected_output)

    def test_extract_measures_middle_measures_to_krn(self):
        options = kp.ExportOptions(spine_types=['**kern'], kern_type=kp.Encoding.normalizedKern, from_measure=2,
                                   to_measure=4)
        importer = kp.Importer()
        document = importer.import_file('resource_dir/legacy/base_tuplet_longer.krn')

        exporter = kp.Exporter()
        real_ouputput = exporter.export_string(document, options)
        with open('resource_dir/legacy/base_tuplet_longer_m2-m4.krn', 'r') as f:
            expected_output = f.read()
        self.assertEqual(real_ouputput, expected_output)

    def test_extract_measures_middle_measures_to_ekrn(self):
        options = kp.ExportOptions(spine_types=['**kern'], kern_type=kp.Encoding.eKern, from_measure=2,
                                   to_measure=4)
        importer = kp.Importer()
        document = importer.import_file('resource_dir/legacy/base_tuplet_longer.krn')

        dot_exporter = kp.GraphvizExporter()
        dot_exporter.export_to_dot(document.tree, f'/tmp/base_tuplet_longer.dot')

        exporter = kp.Exporter()
        actual_output = exporter.export_string(document, options)
        with open('resource_dir/legacy/base_tuplet_longer_m2-m4.ekrn', 'r') as f:
            expected_output = f.read()
        self.assertEqual(actual_output, expected_output)

    def test_extract_measures_bad_measures_input(self):
        importer = kp.Importer()
        document = importer.import_file('resource_dir/legacy/base_tuplet_longer.krn')

        options = kp.ExportOptions(spine_types=['**kern'], from_measure=-1, to_measure=2)
        exporter = kp.Exporter()
        with self.assertRaises(ValueError):
            exporter.export_string(document, options)

        options = kp.ExportOptions(spine_types=['**kern'], from_measure=2, to_measure=-1)
        with self.assertRaises(ValueError):
            exporter.export_string(document, options)

        options = kp.ExportOptions(spine_types=['**kern'], from_measure=3, to_measure=2)
        with self.assertRaises(ValueError):
            exporter.export_string(document, options)

        options = kp.ExportOptions(spine_types=['**kern'], from_measure=None, to_measure=99999)
        with self.assertRaises(ValueError):
            exporter.export_string(document, options)

        options = kp.ExportOptions(spine_types=['**kern'], from_measure=None, to_measure=7)
        with self.assertRaises(ValueError):
            exporter.export_string(document, options)

    def test_extract_measures_when_spines_split_one_spine(self):
        doc, err = kp.load('resource_dir/samples/score_with_dividing_one_spine.krn')

        kp.store_graph(doc, '/tmp/x.dot')
        exported_real = kp.dumps(doc,
                                 spine_types=['**kern'],
                                 encoding=kp.Encoding.normalizedKern,
                                 from_measure=9,
                                 to_measure=13)

        with open('resource_dir/samples/score_with_dividing_one_spine_m9-m13.krn', 'r') as f:
            expected_output = f.read()
        with open('resource_dir/samples/score_with_dividing_one_spine_m9-m13.krn', 'w') as f:
            f.write(exported_real)
        self.assertEqual(expected_output, exported_real)

    def test_extract_measures_when_spines_split_two_spines(self):
        doc, err = kp.read('resource_dir/samples/score_with_dividing_two_spines.krn')
        export_otions = kp.ExportOptions(spine_types=['**kern'],
                                         kern_type=kp.Encoding.normalizedKern,
                                         from_measure=49,
                                         to_measure=56)

        exported_real = kp.export(doc, export_otions)

        with open('resource_dir/samples/score_with_dividing_two_spines_m49-m56.krn', 'r') as f:
            expected_output = f.read()
        self.assertEqual(expected_output, exported_real)

    def testOther(self):
        importer = self.doJustImportTest(
            'resource_dir/polish/test3/pl-wn--sd-xvi-qu-273--001-020_gomolka-mikolaj--melodiae-na-psalterz-polski-xx-wsiadaj-z-dobrym-sercem-o-krolu-cnotliwy.krn')
        print(importer)

    def testStringImporter(self):
        input_kern = "**kern\n4c\n4d\n4e\n4f\n*-"
        importer = kp.Importer()
        document = importer.import_string(input_kern)
        output_kern = kp.dumps(document,
                               spine_types=['**kern'],
                               include=kp.BEKERN_CATEGORIES,
                               encoding=kp.Encoding.eKern, )
        expected_ekern = "**ekern\n4@c\n4@d\n4@e\n4@f\n*-\n"
        self.assertEqual(expected_ekern, output_kern)

    def testLexicalError(self):
        input_kern = "**kern\ncleF4\n4c\n4d\n4e\n4f\n*-"
        importer = kp.Importer()
        importer.import_string(input_kern)
        self.assertEqual(1, len(importer.errors))

    def testParserError_only_the_last(self):
        input_kern = "**kern\n*clefF4\n4d\n4e\n4f\nc4\n*-"
        importer = kp.Importer()
        importer.import_string(input_kern)
        self.assertEqual(1, len(importer.errors))


    def testParserError_only_the_first_ensure_one_error_is_not_being_propagated(self):
        input_kern = "**kern\n*clefF4\nc4\n4d\n4e\n4f\n*-"
        importer = kp.Importer()
        importer.import_string(input_kern)
        self.assertNotEqual(4, len(importer.errors))
        self.assertEqual(1, len(importer.errors))

    def testLexicalParserError(self):
        input_kern = "**kern\ncleF4\nc4\n4d\n4e\n4f\n*-"
        importer = kp.Importer()
        importer.import_string(input_kern)
        self.assertEqual(2, len(importer.errors))

    def testMetacomment(self):
        input_kern = "!!!M1\n**kern\n*clefF4\n4c\n4d\n!!!M2\n4e\n4f\n*-"
        importer = kp.Importer()
        importer.import_string(input_kern)
        self.assertFalse(importer.has_errors())

    def test_metadatacomments_generic(self):
        input_kern_file = 'resource_dir/legacy/chor001.krn'
        output_metadata_array_expected = []
        with open('resource_dir/legacy/chor001-metadata-generic.txt', 'r') as f:
            for line in f:
                output_metadata_array_expected.append(line.strip())

        importer = kp.Importer()
        document = importer.import_file(input_kern_file)
        output_metadata_real = document.get_metacomments()

        self.assertListEqual(output_metadata_array_expected, output_metadata_real)

    def test_metadatacomments_specific_option(self):
        input_kern_file = 'resource_dir/legacy/chor001.krn'
        output_metadata_array_expected = ['!!!COM: Bach, Johann Sebastian']

        importer = kp.Importer()
        document = importer.import_file(input_kern_file)
        output_metadata_array_output = document.get_metacomments(KeyComment='COM')

        self.assertListEqual(output_metadata_array_expected, output_metadata_array_output)  # composer

    def test_metadatacomments_specific_option_clear(self):
        input_kern_file = 'resource_dir/legacy/chor001.krn'
        output_metadata_array_expected = ['Bach, Johann Sebastian']

        importer = kp.Importer()
        document = importer.import_file(input_kern_file)
        output_metadata_array_output = document.get_metacomments(KeyComment='COM', clear=True)

        self.assertListEqual(output_metadata_array_expected, output_metadata_array_output)  # composer

    def has_token(self, tokens, encoding):
        for token in tokens:
            if token.encoding == encoding:
                return True
        return False

    def test_has_token(self):
        input_kern_file = Path('resource_dir/legacy/chor001.krn')
        importer = kp.Importer()
        document = importer.import_file(input_kern_file)
        all_tokens = document.get_all_tokens()
        self.assertTrue(self.has_token(all_tokens, '**kern'))
        self.assertTrue(self.has_token(all_tokens, '*clefF4'))
        self.assertTrue(self.has_token(all_tokens, '4e'))
        self.assertTrue(self.has_token(all_tokens, '8BB'))
        self.assertTrue(self.has_token(all_tokens, '*-'))
        self.assertFalse(self.has_token(all_tokens, '=3'))  # the number is removed from the original file
        self.assertTrue(self.has_token(all_tokens, '='))
        self.assertTrue(self.has_token(all_tokens, '=='))
        self.assertTrue(self.has_token(all_tokens, '.'))
        self.assertTrue(self.has_token(all_tokens, '!tenor'))
        self.assertTrue(self.has_token(all_tokens, '*M3/4'))
        self.assertTrue(self.has_token(all_tokens, '*MM60'))
        self.assertTrue(self.has_token(all_tokens, '*Iorgan'))

    def has_category(self, tokens, category):
        for token in tokens:
            if kp.TokenCategory.is_child(child=token.category, parent=category):
                return True
        return False

    def test_has_token_category(self):
        input_kern_file = Path('resource_dir/legacy/chor001.krn')
        importer = kp.Importer()
        document = importer.import_file(input_kern_file)
        all_tokens = document.get_all_tokens()

        self.assertTrue(self.has_category(all_tokens, kp.TokenCategory.LINE_COMMENTS))
        self.assertTrue(self.has_category(all_tokens, kp.TokenCategory.CORE))
        self.assertTrue(self.has_category(all_tokens, kp.TokenCategory.EMPTY))
        self.assertTrue(self.has_category(all_tokens, kp.TokenCategory.OTHER))
        self.assertTrue(self.has_category(all_tokens, kp.TokenCategory.OTHER_CONTEXTUAL))
        self.assertTrue(self.has_category(all_tokens, kp.TokenCategory.BARLINES))
        self.assertTrue(self.has_category(all_tokens, kp.TokenCategory.SIGNATURES))
        self.assertTrue(self.has_category(all_tokens, kp.TokenCategory.STRUCTURAL))

    def test_get_all_tokens(self):
        # Arrange
        input_kern_file = Path('resource_dir/legacy/chor001.krn')
        importer = kp.Importer()
        document = importer.import_file(input_kern_file)
        expected_tokens = []
        with open('resource_dir/legacy/chor001-all_tokens.txt', 'r') as f:
            for line in f:
                expected_tokens.append(line.strip())

        # Act
        tokens = document.get_all_tokens()
        tokens = [token.encoding for token in tokens]

        # Assert
        self.assertEqual(len(expected_tokens), len(tokens))
        self.assertListEqual(expected_tokens, tokens)

    def test_get_unique_tokens(self):
        # Arrange
        input_kern_file = Path('resource_dir/legacy/chor001.krn')
        importer = kp.Importer()
        document = importer.import_file(input_kern_file)
        with open('resource_dir/legacy/chor001-unique_tokens.txt', 'r') as f:
            expected_tokens = f.read().splitlines()

        # Act
        encodings = document.get_unique_token_encodings()
        encodings.sort()
        expected_tokens.sort()


        # Assert
        self.assertEqual(len(expected_tokens), len(encodings))
        self.assertListEqual(expected_tokens, encodings)

    @unittest.skip("TODO: Add remove measure numbers")
    def test_get_unique_tokens_when_removing_measure_numbers(self):
        # Arrange
        input_kern_file = Path('resource_dir/legacy/chor001.krn')
        importer = kp.Importer()
        document = importer.import_file(input_kern_file)
        with open('resource_dir/legacy/chor001-unique_tokens_without_measure_numbers.txt', 'r') as f:
            expected_tokens = f.read().splitlines()

        # Act
        encodings = document.get_unique_token_encodings()
        encodings.sort()
        expected_tokens.sort()

        # Assert
        self.assertEqual(len(expected_tokens), len(encodings))
        self.assertListEqual(expected_tokens, encodings)

    def test_get_unique_tokens_when_filter_by_categories(self):
        # Arrange
        input_kern_file = Path('resource_dir/legacy/chor001.krn')
        importer = kp.Importer()
        document = importer.import_file(input_kern_file)
        with open('resource_dir/legacy/chor001-unique_tokens_with_category.txt', 'r') as f:
            expected_tokens = f.read().splitlines()

        # Act
        encodings = document.get_unique_token_encodings(
            filter_by_categories=[kp.TokenCategory.CORE, kp.TokenCategory.SIGNATURES])
        encodings.sort()
        expected_tokens.sort()

        # Assert
        self.assertEqual(len(expected_tokens), len(encodings))
        self.assertListEqual(expected_tokens, encodings)

    @unittest.skip("TODO: Instrument is not a valid category yet")
    def test_document_get_voices(self):
        input_kern_file = Path('resource_dir/legacy/chor048.krn')
        expected_voices = ['!bass', '!tenor', '!alto', '!soprno']
        expected_voices_cleaned = ['bass', 'tenor', 'alto', 'soprno']

        importer = kp.Importer()
        document = importer.import_file(input_kern_file)
        voices = document.get_voices()
        voices_cleaned = document.get_voices(clean=True)

        self.assertListEqual(expected_voices, voices)
        self.assertListEqual(expected_voices_cleaned, voices_cleaned)

    @unittest.skip
    def test_voices_range(self):
        pass

    def test_get_kern_from_ekern(self):
        input_ekern = "**ekern\n4@c\n4@d\n4@e\n4@f\n*-"
        expected_kern = "**kern\n4c\n4d\n4e\n4f\n*-"
        real_kern = kp.get_kern_from_ekern(input_ekern)

        self.assertEqual(expected_kern, real_kern)

    def test_document_get_first_measure_ok(self):
        input_kern_file = Path('resource_dir/legacy/chor001.krn')  # This score is correctly formatted
        importer = kp.Importer()
        document = importer.import_file(input_kern_file)
        first_measure = document.get_first_measure()
        self.assertEqual(1, first_measure)

    def test_document_get_first_measure_empty(self):
        importer = kp.Importer()
        document = importer.import_string('**kern\n*-')
        with self.assertRaises(Exception):
            first_measure = document.get_first_measure()

    def test_document_get_last_measure_ok(self):
        input_kern_file = 'resource_dir/legacy/chor048.krn'  # This score has 10 measures
        importer = kp.Importer()
        document = importer.import_file(input_kern_file)
        measures = document.measures_count()
        self.assertEqual(11, measures)

    def test_document_get_last_measure_empty(self):
        importer = kp.Importer()
        document = importer.import_string('**kern\n*-')
        with self.assertRaises(Exception):
            last_measure = document.measures_count()

    def test_document_iterator_basic(self):
        # Arrange
        all_sub_kerns = []
        importer = kp.Importer()
        document = importer.import_file(Path('resource_dir/legacy/chor048.krn'))
        expected_sub_kerns_size = document.measures_count()
        count = 0
        options = kp.ExportOptions(spine_types=['**kern'], kern_type=kp.Encoding.normalizedKern)

        # Act
        for i in range(document.get_first_measure(), document.measures_count() + 1):
            count += 1
            options.from_measure = i
            options.to_measure = i
            exporter = kp.Exporter()
            content = exporter.export_string(document, options)
            print(content)
            all_sub_kerns.append(content)

        # Assert
        self.assertEqual(expected_sub_kerns_size, count)
        # TODO: Check if exported content is correct
        #for kern in all_sub_kerns:
        #    iter_importer = Importer()
        #    iter_document = iter_importer.import_string(kern)
        #    # if the import doesn't raise an exception, the kern is valid
        #    #self.assertFalse(iter_importer.has_errors())

    def test_document_method__iter__(self):
        # Arrange
        all_sub_kerns = []
        all_indexes = []
        importer = kp.Importer()
        document = importer.import_file(Path('resource_dir/legacy/chor048.krn'))
        expected_sub_kerns_size = document.measures_count()

        # Act
        for index in document:
            all_indexes.append(index)
            options = kp.ExportOptions(from_measure=index, to_measure=index, spine_types=['**kern'])
            exporter = kp.Exporter()
            content = exporter.export_string(document, options)
            all_sub_kerns.append(content)

        # Assert
        self.assertEqual(expected_sub_kerns_size, len(all_sub_kerns))

    def test_document_method__next__(self):
        # Arrange
        all_sub_kerns = []
        all_indexes = []
        importer = kp.Importer()
        document = importer.import_file(Path('resource_dir/legacy/chor048.krn'))
        expected_sub_kerns_size = document.measures_count()

        # Act
        iterator = iter(document)
        for index in iterator:
            all_indexes.append(index)
            options = kp.ExportOptions(from_measure=index, to_measure=index, spine_types=['**kern'])
            exporter = kp.Exporter()
            content = exporter.export_string(document, options)
            all_sub_kerns.append(content)

        # Assert
        self.assertEqual(expected_sub_kerns_size, len(all_sub_kerns))

    def test_export_spines(self):
        # Arrange
        importer = kp.Importer()
        document = importer.import_file('resource_dir/legacy/chor048.krn')
        expected_kern_spines = ['**kern', '**kern', '**kern', '**kern']
        expected_all_spines = ['**kern', '**kern', '**kern', '**kern', '**root', '**harm']

        # Act
        exporter = kp.Exporter()
        real_1 = exporter.get_spine_types(document, spine_types=['**kern'])
        real_2 = exporter.get_spine_types(document, spine_types=['**kern', '**root', '**harm'])
        real_3 = exporter.get_spine_types(document, spine_types=None)

        # Assert
        self.assertListEqual(expected_kern_spines, real_1)
        self.assertListEqual(expected_all_spines, real_2)
        self.assertListEqual(expected_all_spines, real_3)

    @unittest.skip("TODO: Complete bug. ISSUE #12 test Eliseo")
    def test_export_string_two_different_spines(self):
        # Arrange
        doc, err = kp.load('resource_dir/legacy/chor048.krn')

        content = kp.dumps(doc, spine_types=['**kern', '**root', '**harm'],
                           encoding=kp.Encoding.eKern)

        doc_2, _ = kp.loads(content)  # raise error if the content is not valid

        print(content)

    def test_never_export_measure_numbers(self):
        # Arrange
        importer = kp.Importer()
        document = importer.import_file(Path('resource_dir/legacy/chor009.krn'))
        options = kp.ExportOptions(spine_types=None,
                                   kern_type=kp.Encoding.eKern,
                                   show_measure_numbers=False)
        exporter = kp.Exporter()
        content = exporter.export_string(document, options)
        print(content)
        measures = range(1, document.measures_count() + 1)
        measures = [f'={m}' for m in
                    measures]  # TODO: David. En los spines que no son kerns siempre se exportan los números de compás
        for measure in measures:
            self.assertNotIn(measure, content)

    def test_always_export_measure_numbers(self):
        # Arrange
        importer = kp.Importer()
        document = importer.import_file(Path('resource_dir/legacy/chor009.krn'))
        options = kp.ExportOptions(spine_types=None,
                                   kern_type=kp.Encoding.eKern,
                                   show_measure_numbers=True)
        exporter = kp.Exporter()
        content = exporter.export_string(document, options)
        print(content)
        measures = range(1, document.measures_count() + 1)
        measures = [f'={m}' for m in measures]
        for measure in measures:
            for line in content.splitlines():  # Find the line with the measure numbers
                if measure in line:
                    tokens = line.split('\t')  # Split the line in tokens
                    for token in tokens:
                        self.assertEqual(measure, token)

    def test_kern_with_first_measure_is_exported_without_the_measure_number_in_all_spines(self):
        # Arrange
        input_kern_file = 'resource_dir/samples/haydn-sonate-15_1-original.krn'
        output_kern_file = 'resource_dir/samples/haydn-sonate-15_1-output.krn'
        with open(output_kern_file, 'r') as f:
            expected_output = f.read()

        # Act
        doc, err = kp.load(input_kern_file)
        real_output = kp.dumps(doc)
        kp.graph(doc, '/tmp/graph.dot')

        # Assert
        self.assertEqual(expected_output, real_output)


if __name__ == '__main__':
    unittest.main()



import unittest
import logging

import kernpy as kp
import kernpy as kp


logger = logging.getLogger(__name__)


class TestGKernExporter(unittest.TestCase):

    def setUp(self):
        self.gclef = kp.GClef()
        self.staff = kp.Staff()
        self.exporter = kp.GKernExporter(self.gclef)

    def assertExport(self, pitch: kp.AgnosticPitch, expected_str: str):
        self.assertEqual(self.exporter.export(self.staff, pitch), expected_str)

    def assertAgnosticPosition(self, pitch: kp.AgnosticPitch, expected_pos: kp.PositionInStaff):
        pos = self.exporter.agnostic_position(self.staff, pitch)
        self.assertEqual(pos, expected_pos)

    def test_init_stores_clef(self):
        self.assertEqual(self.exporter.clef, self.gclef)

    def test_export_middle_E4(self):
        pitch = kp.AgnosticPitch('E', 4)
        expected = f"T@1"
        self.assertExport(pitch, expected)

    def test_export_with_accidentals(self):
        self.assertExport(kp.AgnosticPitch('E+', 4), f"T@1")
        self.assertExport(kp.AgnosticPitch('E-', 4), f"T@1")

    def test_export_scale_up(self):
        pitches_positions = [
            ('F', 4, 'S@1'),
            ('G', 4, 'T@2'),
            ('A', 4, 'S@2'),
            ('B', 4, 'T@3'),
            ('C', 5, 'S@3'),
            ('D', 5, 'T@4'),
            ('E', 5, 'S@4'),
            ('F', 5, 'T@5'),
            ('G', 5, 'S@5'),
        ]
        for name, octave, pos_str in pitches_positions:
            with self.subTest(pitch=f"{name}{octave}"):
                pitch = kp.AgnosticPitch(name, octave)
                expected = f"{pos_str}"
                self.assertExport(pitch, expected)

    def test_export_scale_down(self):
        pitches_positions = [
            ('D', 4, 'S@0'),
            ('C', 4, 'T@0'),
            ('B', 3, 'S@-1'),
            ('A', 3, 'T@-1'),
            ('G', 3, 'S@-2'),
            ('F', 3, 'T@-2'),
            ('E', 3, 'S@-3'),
        ]
        for name, octave, pos_str in pitches_positions:
            with self.subTest(pitch=f"{name}{octave}"):
                pitch = kp.AgnosticPitch(name, octave)
                expected = f"{pos_str}"
                self.assertExport(pitch, expected)

    def test_export_multiple_clefs(self):
        clefs = [
            (kp.GClef(), kp.AgnosticPitch('E', 4), 'T@1'),
            (kp.F3Clef(), kp.AgnosticPitch('B', 3), 'T@1'),
            (kp.F4Clef(), kp.AgnosticPitch('G', 2), 'T@1'),
            (kp.C1Clef(), kp.AgnosticPitch('C', 3), 'T@1'),
            (kp.C2Clef(), kp.AgnosticPitch('A', 2), 'T@1'),
            (kp.C3Clef(), kp.AgnosticPitch('B', 2), 'T@1'),
            (kp.C4Clef(), kp.AgnosticPitch('D', 2), 'T@1'),
        ]
        for clef, pitch, pos_str in clefs:
            with self.subTest(clef=clef, pitch=pitch):
                exporter = kp.GKernExporter(clef)
                result = exporter.export(self.staff, pitch)
                self.assertEqual(result, f"{pos_str}")

    def test_agnostic_position_equivalence(self):
        pitch = kp.AgnosticPitch('F', 4)
        expected_pos = kp.PositionInStaff.from_space(1)
        self.assertAgnosticPosition(pitch, expected_pos)

    def test_export_with_large_range(self):
        pitches = [
            kp.AgnosticPitch('C', 2),
            kp.AgnosticPitch('E', 2),
            kp.AgnosticPitch('G', 2),
            kp.AgnosticPitch('B', 5),
            kp.AgnosticPitch('C', 6),
            kp.AgnosticPitch('D', 6),
        ]
        for pitch in pitches:
            with self.subTest(pitch=pitch):
                result = self.exporter.export(self.staff, pitch)
                self.assertRegex(result, rf"(T|S)@-?\d+")

    def test_export_pitch_with_double_sharps(self):
        self.assertExport(kp.AgnosticPitch('F++', 4), f"S@1")

    def test_export_pitch_with_double_flats(self):
        self.assertExport(kp.AgnosticPitch('F--', 4), f"S@1")


class TestPitchToGKernString(unittest.TestCase):
    def assertPitchToGKern(self, name: str, octave: int, clef: kp.Clef, expected: str):
        pitch = kp.AgnosticPitch(name, octave)
        self.assertEqual(kp.pitch_to_gkern_string(pitch, clef), expected)

    def test_basic_conversion(self):
        self.assertPitchToGKern('E', 4, kp.GClef(), 'T@1')

    def test_conversion_multiple_clefs(self):
        test_cases = [
            ('E', 4, kp.GClef(), 'T@1'),
            ('B', 3, kp.F3Clef(), 'T@1'),
            ('C', 3, kp.C1Clef(), 'T@1'),
        ]
        for name, octave, clef, expected in test_cases:
            with self.subTest(pitch=f"{name}{octave}", clef=clef):
                self.assertPitchToGKern(name, octave, clef, expected)

    def test_with_accidentals_flat(self):
        self.assertPitchToGKern('A-', 4, kp.GClef(), 'S@2')

    def test_with_accidentals_sharp(self):
        self.assertPitchToGKern('C+', 5, kp.GClef(), 'S@3')

    def test_pitch_extremes(self):
        extremes = [
            ('C', 2, kp.C2Clef()),
            ('B', 5, kp.GClef()),
        ]
        for name, octave, clef in extremes:
            with self.subTest(pitch=f"{name}{octave}", clef=clef):
                result = kp.pitch_to_gkern_string(kp.AgnosticPitch(name, octave), clef)
                self.assertRegex(result, rf"(T|S)@-?\d+")

    def test_same_pitch_different_clefs(self):
        pitch = kp.AgnosticPitch('D', 4)
        clefs_position_representation = [
            (kp.GClef(), f'S@0'),
            (kp.F3Clef(), f'T@2'),
            (kp.F4Clef(), f'S@6'),
            (kp.C1Clef(), f'T@5'),
            (kp.C2Clef(), f'T@6'),
            (kp.C3Clef(), f'S@5'),
            (kp.C4Clef(), f'T@8'),
        ]

        for clef, expected in clefs_position_representation:
            result = kp.pitch_to_gkern_string(pitch, clef)
            logger.info(f"Testing {pitch} with {clef}:\Expected: {expected}.\nGot: {result}")
            self.assertEqual(result, expected)

    def test_gkern_to_g_clef_pitch_basic(self):
        # Around middle C: T@0 == 'c', S@0 == 'd', T@1 == 'e'
        cases = [
            ('T@0', 'c'),
            ('S@0', 'd'),
            ('T@1', 'e'),
        ]
        for token, expected in cases:
            with self.subTest(token=token):
                self.assertEqual(kp.gkern_to_g_clef_pitch(token), expected)

    def test_gkern_to_g_clef_pitch_mixed_examples(self):
        # Mix of above and below middle C, including octave changes
        cases = [
            ('S@3', 'cc'),   # up to c5
            ('T@4', 'dd'),   # up to d5
            ('S@-2', 'G'),   # down to G3
            ('T@-4', 'BB'),  # down to B2 (two-letter uppercase)
        ]
        for token, expected in cases:
            with self.subTest(token=token):
                self.assertEqual(kp.gkern_to_g_clef_pitch(token), expected)

    def test_gkern_to_g_clef_pitch_bad_inputs(self):
        bad_tokens = ['X@1', 'T@foo', 'bogus']
        for token in bad_tokens:
            with self.subTest(token=token):
                with self.assertRaises(ValueError):
                    kp.gkern_to_g_clef_pitch(token)

    def test_import_from_standard_to_agnostic_and_export_back(self):
        source_key = kp.GClef()
        source_pitches = {
            'ccc',
            'bb',
            'aa',
            'gg',
            'ff',
            'ee',
            'dd',
            'cc',
            'b',
            'a',
            'b',
            'f',
            'e',
            'd',
            'c',
            'B',
            'A',
            'G',
            'F',
            'E',
            'D',
            'C',
            'BB',
        }

        for g_clef_pitch in source_pitches:
            with self.subTest(g_clef_pitch=g_clef_pitch):
                pitch_importer = kp.PitchImporterFactory.create('kern')
                agnostic_pitch: kp.AgnosticPitch = pitch_importer.import_pitch(g_clef_pitch)
                agnostic_encoding: str = kp.pitch_to_gkern_string(
                    agnostic_pitch,
                    source_key
                )
                self.assertEqual(agnostic_encoding, kp.pitch_to_gkern_string(agnostic_pitch, source_key))



if __name__ == "__main__":
    unittest.main()



import unittest
import os
from pathlib import Path

import kernpy as kp


class ImporterTestCase(unittest.TestCase):

    def test_import_from_file(self):
        input_kern_file = 'resource_dir/legacy/chor001.krn'
        importer = kp.Importer()
        document = importer.import_file(input_kern_file)
        self.assertIsNotNone(document)

    def test_import_from_string(self):
        input_kern_file = 'resource_dir/legacy/chor001.krn'
        with open(input_kern_file, 'r', newline='', encoding='utf-8', errors='ignore') as f:
            content = f.read()
        importer = kp.Importer()
        document = importer.import_string(content)
        self.assertIsNotNone(document)

    def test_spine_ids(self):
        input_kern_file = 'resource_dir/legacy/chor001.krn'

        doc, err = kp.read(input_kern_file)
        header_tokens = doc.get_all_tokens()
        spines_ids = [t.spine_id for t in header_tokens if isinstance(t, kp.core.HeaderToken)]
        self.assertListEqual([0, 1, 2, 3, 4, 5], spines_ids)

    def test_not_raise_handled_exception_when_can_bad_header(self):
        input_kern_file = 'resource_dir/samples/wrong_header.krn'
        doc, err = kp.load(input_kern_file)

        self.assertTrue(True, "Should not raise an exception when handling a bad header")

    def test_raise_wrong_number_of_columns(self):
        input_kern_file = 'resource_dir/samples/wrong_number_of_columns.krn'
        with self.assertRaises(ValueError):
            doc, err = kp.load(input_kern_file)

    @unittest.skip
    def test_fix_error_wrong_number_of_columns(self):
        # Arrange
        input_kern_file = 'resource_dir/samples/wrong_number_of_columns.krn'
        output_kern_file = 'resource_dir/samples/wrong_number_of_columns_fixed.krn'
        with open(output_kern_file, 'r') as f:
            expected_content = f.read()

        # Act
        doc, err = kp.read(input_kern_file)
        real_content = kp.export(doc, kp.ExportOptions())

        # Assert
        self.assertEqual(expected_content, real_content)

    def test_has_instrument_tokens(self):
        input_kern_file = 'resource_dir/legacy/chor001.krn'
        doc, err = kp.read(input_kern_file)
        self.assertTrue(len(doc.get_all_tokens(filter_by_categories=[kp.TokenCategory.INSTRUMENTS])) > 0)

    def test_import_unexpected_header_is_ok(self):
        input_kern_file = 'resource_dir/samples/unexpected_header.krn'

        doc, err = kp.load(input_kern_file)
        self.assertIsNotNone(doc, "Document should not be None")

    def test_any_header_found(self):
        input_kern_file = Path('resource_dir/samples/any_header.krn')
        expected_message = (
f'Any spine header found in the column #0. '
f'Expected a previous line with valid content. The token in column #0 and row '
f'#0 was not created correctly. Error detected in column #0 in row #1. Found '
f'*ClefF2. '
)

        with self.assertRaises(ValueError) as context:
            doc, _ = kp.load(input_kern_file)

        # Assert that the error message is as expected
        self.assertEqual(str(context.exception), expected_message)




"""

This module is responsible for generating the Graphic **kern encoding.

The Graphic **kern encoding (**gkern) is an agnostify **kern encoding. It keeps the original **kern encoding structure \
but replaces the pitches with its graphic representation. So the E pitch in G Clef will use the same graphic \
representation as the C pitch in C in 1st Clef or the A in F in 4th Clef.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from enum import Enum
from typing import Dict

from .tokens import (
    TOKEN_SEPARATOR, ClefToken, Token, TokenCategory
)

from .transposer import (
    transpose,
    Direction,
    distance,
    agnostic_distance,
    transpose_agnostics,
    transpose_encoding_to_agnostic,
    transpose_agnostic_to_encoding,
)

from .pitch_models import (
    AgnosticPitch,
    pitches,
)

LETTERS = ['c', 'd', 'e', 'f', 'g', 'a', 'b']

class Alteration(Enum):
    """
    Enum for the alteration of a pitch.
    """
    NONE = 0
    SHARP = 1
    FLAT = -1
    DOUBLE_SHARP = 2
    DOUBLE_FLAT = -2
    TRIPLE_SHARP = 3
    TRIPLE_FLAT = -3
    HALF_SHARP = 0.5
    HALF_FLAT = -0.5
    QUARTER_SHARP = 0.25
    QUARTER_FLAT = -0.25

    def __str__(self) -> str:
        return self.name


class PositionInStaff:
    """
    The pair S/T was chosen for clarity and coherence:
    S directly abbreviates Space (as in “Space bar”),
    while T evokes Trace, Track, or Tier, terms associated with lines in English, German, and French.
    """
    LINE_CHARACTER = 'T'
    SPACE_CHARACTER = 'S'

    def __init__(self, line_space: int):
        """
        Initializes the PositionInStaff object.

        Args:
            line_space (int): 0 for bottom line, -1 for space under bottom line, 1 for space above bottom line. \
             Increments by 1 for each line or space.

        """
        self.line_space = line_space

    @classmethod
    def from_line(cls, line: int) -> PositionInStaff:
        """
        Creates a PositionInStaff object from a line number.

        Args:
            line (int): The line number. line 1 is bottom line, 2 is the 1st line from bottom, 0 is the bottom ledger line

        Returns:
            PositionInStaff: The PositionInStaff object. 0 for the bottom line, 2 for the 1st line from bottom, -1 for the bottom ledger line, etc.
        """
        return cls((line - 1) * 2)

    @classmethod
    def from_space(cls, space: int) -> PositionInStaff:
        """
        Creates a PositionInStaff object from a space number.

        Args:
            space (int): The space number. space 1 is bottom space, 2

        Returns:
            PositionInStaff: The PositionInStaff object.
        """
        return cls((space) * 2 - 1)

    @classmethod
    def from_encoded(cls, encoded: str) -> PositionInStaff:
        """
        Creates a PositionInStaff object from an encoded string.

        Args:
            encoded (str): The encoded string.

        Returns:
            PositionInStaff: The PositionInStaff object.
        """
        if encoded.startswith(cls.LINE_CHARACTER):
            line = int(encoded[1:])  # Extract the line number
            return cls.from_line(line)
        elif encoded.startswith(cls.SPACE_CHARACTER):
            space = int(encoded[1:])  # Extract the space number
            return cls.from_space(space)
        else:
            raise ValueError(f""
                             f"Invalid encoded string: {encoded}. "
                             f"Expected to start with '{cls.LINE_CHARACTER}' or '{cls.SPACE_CHARACTER} at the beginning.")


    def line(self):
        """
        Returns the line number of the position in staff.
        """
        return self.line_space // 2 + 1


    def space(self):
        """
        Returns the space number of the position in staff.
        """
        return (self.line_space - 1) // 2 + 1


    def is_line(self) -> bool:
        """
        Returns True if the position is a line, False otherwise. If is not a line, it is a space, and vice versa.
        """
        return self.line_space % 2 == 0

    def move(self, line_space_difference: int) -> PositionInStaff:
        """
        Returns a new PositionInStaff object with the position moved by the given number of lines or spaces.

        Args:
            line_space_difference (int): The number of lines or spaces to move.

        Returns:
            PositionInStaff: The new PositionInStaff object.
        """
        return PositionInStaff(self.line_space + line_space_difference)

    def position_below(self) -> PositionInStaff:
        """
        Returns the position below the current position.
        """
        return self.move(-2)

    def position_above(self) -> PositionInStaff:
        """
        Returns the position above the current position.
        """
        return self.move(2)



    def __str__(self) -> str:
        """
        Returns the string representation of the position in staff.
        """
        if self.is_line():
            return f"{self.LINE_CHARACTER}{TOKEN_SEPARATOR}{int(self.line())}"
        else:
            return f"{self.SPACE_CHARACTER}{TOKEN_SEPARATOR}{int(self.space())}"

    def __repr__(self) -> str:
        """
        Returns the string representation of the PositionInStaff object.
        """
        return f"PositionInStaff(line_space={self.line_space}), {self.__str__()}"

    def __eq__(self, other) -> bool:
        """
        Compares two PositionInStaff objects.
        """
        if not isinstance(other, PositionInStaff):
            return False
        return self.line_space == other.line_space

    def __ne__(self, other) -> bool:
        """
        Compares two PositionInStaff objects.
        """
        return not self.__eq__(other)

    def __hash__(self) -> int:
        """
        Returns the hash of the PositionInStaff object.
        """
        return hash(self.line_space)

    def __lt__(self, other) -> bool:
        """
        Compares two PositionInStaff objects.
        """
        if not isinstance(other, PositionInStaff):
            return NotImplemented
        return self.line_space < other.line_space


class DiatonicPitch:
    def __init__(self, diatonic_pitch: str) -> None:
        if diatonic_pitch not in pitches:
            raise ValueError(f"Invalid diatonic pitch: {diatonic_pitch}. "
                             f"Expected one of {pitches}.")
        self.encoding = diatonic_pitch

    def __str__(self) -> str:
        return self.encoding


class PitchPositionReferenceSystem:
    def __init__(self, base_pitch: AgnosticPitch):
        """
        Initializes the PitchPositionReferenceSystem object.
        Args:
            base_pitch (AgnosticPitch): The AgnosticPitch in the first line of the Staff. \
             The AgnosticPitch object that serves as the reference point for the system.
        """
        self.base_pitch = base_pitch

    def compute_position(self, pitch: AgnosticPitch) -> PositionInStaff:
        """
        Computes the position in staff for the given pitch.
        Args:
            pitch (AgnosticPitch): The AgnosticPitch object to compute the position for.
        Returns:
            PositionInStaff: The PositionInStaff object representing the computed position.
        """
        # map A–G to 0–6
        LETTER_TO_INDEX = {'C': 0, 'D': 1, 'E': 2,
                           'F': 3, 'G': 4, 'A': 5, 'B': 6}

        # strip off any '+' or '-' accidentals, then grab the letter
        def letter(p: AgnosticPitch) -> str:
            name = p.name.replace('+', '').replace('-', '')
            return AgnosticPitch(name, p.octave).name

        base_letter_idx = LETTER_TO_INDEX[letter(self.base_pitch)]
        target_letter_idx = LETTER_TO_INDEX[letter(pitch)]

        # "octave difference × 7" plus the letter‐index difference
        diatonic_steps = (pitch.octave - self.base_pitch.octave) * 7 \
                         + (target_letter_idx - base_letter_idx)

        # that many "lines or spaces" above (or below) the reference line
        return PositionInStaff(diatonic_steps)




# TARGET: Retrieve agnostic encoding from semantic encoding
# INPUT: Pitch (Semantic) + Clef (Semantic)
# OUTPUT: PositionInStaff (Agnostic)

# CLEF has
# 1. bottom line equivalent to C4
# - C in 1st clef: 1st line is F2 == F
# - C in 2nd clef: 1st line is D2 == D
# - C in 3rd clef: 1st line is B2 == BB
# - C in 4th clef: 1st line is G2 == GG
# - G clef: 1st line is E4 == e
# - F in 3rd clef: 1st line is B2 == BB
# - F in 4th clef: 1st line is G2 == GG

# New abstraction:
# TARGET: PositionInStaff
# INPUT: Pitch (Semantic) + Reference System to the first line
# OUTPUT: PositionInStaff (Agnostic)
# Example:
# Input: Pitch (A4) + Reference System (Line 1 == E3) (Case of G Clef)
# >>> A4 is a 4th minor above E3   ¿How to calculate this?
# >>> a 4th minor is equivalent to the 2nd upper space
# >>> PositionInStaff.from_space(2)
# >>> PositionInStaff(?)

class Clef(ABC):
    """
    Abstract class representing a clef.
    """

    def __init__(self, diatonic_pitch: DiatonicPitch, on_line: int):
        """
        Initializes the Clef object.
        Args:
            diatonic_pitch (DiatonicPitch): The diatonic pitch of the clef (e.g., 'C', 'G', 'F'). This value is used as a decorator.
            on_line (int): The line number on which the clef is placed (1 for bottom line, 2 for 1st line from bottom, etc.). This value is used as a decorator.
        """
        self.diatonic_pitch = diatonic_pitch
        self.on_line = on_line

    @abstractmethod
    def bottom_line(self) -> AgnosticPitch:
        """
        Returns the pitch of the bottom line of the staff.
        """
        ...

    def name(self):
        """
        Returns the name of the clef.
        """
        return f"{self.diatonic_pitch} on line {self.on_line}"

    def reference_point(self) -> PitchPositionReferenceSystem:
        """
        Returns the reference point for the clef.
        """
        return PitchPositionReferenceSystem(self.bottom_line())

    def __str__(self) -> str:
        """
        Returns:
            str: The string representation of the clef.
        """
        return f'{self.diatonic_pitch.encoding.upper()} on the {self.on_line}{self._ordinal_suffix(self.on_line)} line'

    @staticmethod
    def _ordinal_suffix(number: int) -> str:
        """
        Returns the ordinal suffix for a given integer (e.g. 'st', 'nd', 'rd', 'th').

        Args:
            number (int): The number to get the suffix for.

        Returns:
            str: The ordinal suffix.
        """
        # 11, 12, 13 always take “th”
        if 11 <= (number % 100) <= 13:
            return 'th'
        # otherwise use last digit
        last = number % 10
        if last == 1:
            return 'st'
        elif last == 2:
            return 'nd'
        elif last == 3:
            return 'rd'
        else:
            return 'th'


class GClef(Clef):
    def __init__(self):
        """
        Initializes the G Clef object.
        """
        super().__init__(DiatonicPitch('G'), 2)

    def bottom_line(self) -> AgnosticPitch:
        """
        Returns the pitch of the bottom line of the staff.
        """
        return AgnosticPitch('E', 4)

class F3Clef(Clef):
    def __init__(self):
        """
        Initializes the F Clef object.
        """
        super().__init__(DiatonicPitch('F'), 3)

    def bottom_line(self) -> AgnosticPitch:
        """
        Returns the pitch of the bottom line of the staff.
        """
        return AgnosticPitch('B', 3)

class F4Clef(Clef):
    def __init__(self):
        """
        Initializes the F Clef object.
        """
        super().__init__(DiatonicPitch('F'), 4)

    def bottom_line(self) -> AgnosticPitch:
        """
        Returns the pitch of the bottom line of the staff.
        """
        return AgnosticPitch('G', 2)

class C1Clef(Clef):
    def __init__(self):
        """
        Initializes the C Clef object.
        """
        super().__init__(DiatonicPitch('C'), 1)

    def bottom_line(self) -> AgnosticPitch:
        """
        Returns the pitch of the bottom line of the staff.
        """
        return AgnosticPitch('C', 3)

class C2Clef(Clef):
    def __init__(self):
        """
        Initializes the C Clef object.
        """
        super().__init__(DiatonicPitch('A'), 2)

    def bottom_line(self) -> AgnosticPitch:
        """
        Returns the pitch of the bottom line of the staff.
        """
        return AgnosticPitch('A', 2)


class C3Clef(Clef):
    def __init__(self):
        """
        Initializes the C Clef object.
        """
        super().__init__(DiatonicPitch('C'), 3)

    def bottom_line(self) -> AgnosticPitch:
        """
        Returns the pitch of the bottom line of the staff.
        """
        return AgnosticPitch('B', 2)

class C4Clef(Clef):
    def __init__(self):
        """
        Initializes the C Clef object.
        """
        super().__init__(DiatonicPitch('C'), 4)

    def bottom_line(self) -> AgnosticPitch:
        """
        Returns the pitch of the bottom line of the staff.
        """
        return AgnosticPitch('D', 2)


class ClefFactory:
    CLEF_NAMES = { 'G', 'F', 'C' }
    @classmethod
    def create_clef(cls, encoding: str) -> Clef:
        """
        Creates a Clef object based on the given token.

        Clefs are encoded in interpretation tokens that start with a single * followed by the string clef and then the shape and line position of the clef. For example, a treble clef is *clefG2, with G meaning a G-clef, and 2 meaning that the clef is centered on the second line up from the bottom of the staff. The bass clef is *clefF4 since it is an F-clef on the fourth line of the staff.
        A vocal tenor clef is represented by *clefGv2, where the v means the music should be played an octave lower than the regular clef’s sounding pitches. Try creating a vocal tenor clef in the above interactive example. The v operator also works on the other clefs (but these sorts of clefs are very rare). Another rare clef is *clefG^2 which is the opposite of *clefGv2, where the music is written an octave lower than actually sounding pitch for the normal form of the clef. You can also try to create exotic two-octave clefs by doubling the ^^ and vv markers.

        Args:
            encoding (str): The encoding of the clef token.

        Returns:

        """
        encoding = encoding.replace('*clef', '')

        # at this point the encoding is like G2, F4,... or Gv2, F^4,... or G^^2, Fvv4,... or G^^...^^2, Fvvv4,...
        name = list(filter(lambda x: x in cls.CLEF_NAMES, encoding))[0]
        line = int(list(filter(lambda x: x.isdigit(), encoding))[0])
        decorators = ''.join(filter(lambda x: x in ['^', 'v'], encoding))

        if name not in cls.CLEF_NAMES:
            raise ValueError(f"Invalid clef name: {name}. Expected one of {cls.CLEF_NAMES}.")

        if name == 'G':
            return GClef()
        elif name == 'F':
            if line == 3:
                return F3Clef()
            elif line == 4:
                return F4Clef()
            else:
                raise ValueError(f"Invalid F clef line: {line}. Expected 3 or 4.")
        elif name == 'C':
            if line == 1:
                return C1Clef()
            elif line == 2:
                return C2Clef()
            elif line == 3:
                return C3Clef()
            elif line == 4:
                return C4Clef()
            else:
                raise ValueError(f"Invalid C clef line: {line}. Expected 1, 2, 3 or 4.")
        else:
            raise ValueError(f"Invalid clef name: {name}. Expected one of {cls.CLEF_NAMES}.")






class Staff:
    def position_in_staff(self, *, clef: Clef, pitch: AgnosticPitch) -> PositionInStaff:
        """
       Computes the agnostic position in staff for the given clef and pitch.

       Args:
           clef (Clef): The clef defining the reference.
           pitch (AgnosticPitch): The pitch to locate.

       Returns:
           PositionInStaff: The position of the pitch relative to the staff.
       """
        reference = clef.reference_point()
        return reference.compute_position(pitch)



class GKernExporter:
    def __init__(self, clef: Clef):
        self.clef = clef

    def export(self, staff: Staff, pitch: AgnosticPitch) -> str:
        """
        Exports the given pitch to a graphic **kern encoding.
        """
        position = self.agnostic_position(staff, pitch)
        return str(position)

    def agnostic_position(self, staff: Staff, pitch: AgnosticPitch) -> PositionInStaff:
        """
        Returns the agnostic position in staff for the given pitch.
        """
        return staff.position_in_staff(clef=self.clef, pitch=pitch)


def gkern_to_g_clef_pitch(gkern_content: str) -> str:
    """
    Convert a graphic staff position like 'T@4' / 'S@-3' to a **kern pitch (no accidentals),
    using the (G2) clef with middle C = 'c' == T@0.
    (Zero positions exist: T@0 -> 'c', S@0 -> 'd')

    Args:
        gkern_content (str): The graphic staff position string ('T@N' or 'S@N').

    Returns:
        str: only the pitch: 'c', 'dd', 'B', 'BB', ...
    """
    token = gkern_content.strip()
    try:
        position_type, raw_num = token.split(TOKEN_SEPARATOR, 1)
    except ValueError:
        raise ValueError(f"Bad token {token!r}. Expect 'T@N' or 'S@N'.")

    if position_type not in ('T', 'S'):
        raise ValueError(f"Unknown position type {position_type!r}; use 'T' (line) or 'S' (space).")

    try:
        n = int(raw_num)
    except ValueError:
        raise ValueError(f"Bad position number {raw_num!r}; must be an integer.")

    # Diatonic distance from middle C ('c'), anchored at T@0 = 0.
    # Each staff step (line/space) is one diatonic step.
    # Lines are even steps; spaces are odd steps relative to T@0.
    distance = 2 * n + (1 if position_type == 'S' else 0)

    # Map diatonic distance to **kern pitch letters
    idx = distance % 7
    octs = distance // 7

    if distance > 0:
        return LETTERS[idx] * (octs + 1)      # lowercase for c and above
    if distance < 0:
        return LETTERS[idx].upper() * (-octs) # uppercase for below c
    return 'c'


def pitch_to_gkern_string(pitch: AgnosticPitch, clef: Clef) -> str:
    """
    Converts a given pitch and clef to a graphic **kern string representation.

    Args:
        pitch (AgnosticPitch): The pitch to convert.
        clef (Clef): The clef defining the staff context.

    Returns:
        str: The graphic **kern representation (e.g., '|L2').
    """
    staff = Staff()
    exporter = GKernExporter(clef)
    gkern_encoding = exporter.export(staff, pitch) # (e.g., 'L@2', S@-1', etc.)

    accidentals = pitch.accidentals()
    return gkern_to_g_clef_pitch(gkern_encoding) + accidentals  # a nornal **kern pitch with accidentals in G clef



import requests
from PIL import Image
from io import BytesIO
import os


def download_and_save_image(url, save_path):
    try:
        # Make a GET request to the URL
        response = requests.get(url)
        response.raise_for_status()  # Raise an exception for bad responses

        # Open the image using Pillow
        image = Image.open(BytesIO(response.content))

        # Save the image to the specified path
        image.save(save_path, format='JPEG')

        print(f"Image downloaded and saved to: {save_path}")
    except Exception as e:
        print(f"An error occurred: {e}")


def get_image_urls(_manifest_url):
    # It returns the URL of pages tagged with a page number
    # It returns a map with the page label as key, and the URL as
    response = requests.get(_manifest_url)
    manifest_data = response.json()

    items = manifest_data.get('items', [])
    print(f'Items: ', len(items))

    for item in items:
        pl = item.get('label').get('pl')
        if pl:
            page_num = pl[0]
            if page_num != '[]':
                url = item.get('items')[0].get('items')[0].get('id', '')
                print(f'Page #{page_num}, {url}')
                #image_path = os.path.join(output_folder, page_num + '.jpg')
                #download_and_save_image(url, image_path)


if __name__ == "__main__":
    # Replace the manifest_url with the actual URL of your IIIF manifest
    manifest_url = "https://polona2.pl/iiif/item/MTk4NjI5Mw/manifest.json"
    get_image_urls(manifest_url)



"""
Public API for KernPy.

The main functions for handling the input and output of **kern files are provided here.
"""
from __future__ import annotations

from pathlib import Path
from typing import List, Optional, Any, Union, Tuple
from collections.abc import Sequence

from kernpy.core import Importer, Document, Exporter, ExportOptions, GraphvizExporter, TokenCategoryHierarchyMapper
from kernpy.core._io import _write
from kernpy.util.helpers import deprecated


class Generic:
    """
    Generic class.

    This class provides support to the public API for KernPy.

    The main functions implementation are provided here.
    """

    @classmethod
    def read(
            cls,
            path: Path,
            strict: Optional[bool] = False
    ) -> (Document, List[str]):
        """

        Args:
            path:
            strict:

        Returns:

        """
        importer = Importer()
        document = importer.import_file(path)
        errors = importer.errors

        if strict and len(errors) > 0:
            raise Exception(importer.get_error_messages())

        return document, errors

    @classmethod
    def create(
            cls,
            content: str,
            strict: Optional[bool] = False
    ) -> (Document, List[str]):
        """

        Args:
            content:
            strict:

        Returns:

        """
        importer = Importer()
        document = importer.import_string(content)
        errors = importer.errors

        if strict and len(errors) > 0:
            raise Exception(importer.get_error_messages())

        return document, errors

    @classmethod
    def export(
            cls,
            document: Document,
            options: ExportOptions
    ) -> str:
        """

        Args:
            document:
            options:

        Returns:

        """
        exporter = Exporter()
        return exporter.export_string(document, options)

    @classmethod
    def store(
            cls,
            document: Document,
            path: Path,
            options: ExportOptions
    ) -> None:
        """

        Args:
            document:
            path:
            options:

        Returns:
        """
        content = cls.export(document, options)
        _write(path, content)

    @classmethod
    def store_graph(
            cls,
            document: Document,
            path: Path
    ) -> None:
        """

        Args:
            document:
            path:

        Returns:
        """
        graph_exporter = GraphvizExporter()
        graph_exporter.export_to_dot(document.tree, path)

    @classmethod
    def get_spine_types(
            cls,
            document: Document,
            spine_types: Optional[Sequence[str]] = None
    ) -> List[str]:
        """

        Args:
            document:
            spine_types:

        Returns:

        """
        exporter = Exporter()
        return exporter.get_spine_types(document, spine_types)

    @classmethod
    def merge(
            cls,
            contents: Sequence[str],
            strict: Optional[bool] = False
    ) -> Tuple[Document, List[Tuple[int, int]]]:
        """

        Args:
            contents:
            strict:

        Returns:

        """
        if len(contents) < 2:
            raise ValueError(f"Concatenation action requires at least two documents to concatenate."
                             f"But {len(contents)} was given.")

        raise NotImplementedError("The merge function is not implemented yet.")

        doc_a, err_a = cls.create(contents[0], strict=strict)
        for i, content in enumerate(contents[1:]):
            doc_b, err_b = cls.create(content, strict=strict)

            if strict and (len(err_a) > 0 or len(err_b) > 0):
                raise Exception(f"Errors were found during the creation of the documents "
                                f"while using the strict=True option. "
                                f"Description: concatenating: {err_a if len(err_a) > 0 else err_b}")

            doc_a.add(doc_b)
        return cls.export(
            document=doc_a,
            options=options
        )

    @classmethod
    def concat(
            cls,
            contents: Sequence[str],
            separator: Optional[str] = None
    ) -> Tuple[Document, List[Tuple[int, int]]]:
        """

        Args:
            contents:
            separator:

        Returns:

        """
        # Raw kern content
        if separator is None:
            separator = '\n'

        if len(contents) == 0:
            raise ValueError("No contents to merge. At least one content is required.")

        raw_kern = ''
        document = None
        indexes = []
        low_index = 0
        high_index = 0

        # Merge all fragments
        for content in contents:
            raw_kern += separator + content
            document, _ = create(raw_kern)
            high_index = document.measures_count()
            indexes.append((low_index, high_index))

            low_index = high_index + 1  # Next fragment start is the previous fragment end + 1

        if document is None:
            raise Exception("Failed to merge the contents. The document is None.")

        return document, indexes

    @classmethod
    def parse_options_to_ExportOptions(
            cls,
            **kwargs: Any
    ) -> ExportOptions:
        """

        Args:
            **kwargs:

        Returns:

        """
        options = ExportOptions.default()

        # Compute the valid token categories
        options.token_categories = TokenCategoryHierarchyMapper.valid(
            include=kwargs.get('include', None),
            exclude=kwargs.get('exclude', None)
        )

        # Use kwargs to update the ExportOptions object
        for key, value in kwargs.items():
            if key in ['include', 'exclude', 'token_categories']:  # Skip these keys: generated manually
                continue

            if value is not None:
                setattr(options, key, value)

        return options


@deprecated("Use 'load' instead.")
def read(
        path: Union[str, Path],
        strict: Optional[bool] = False
) -> (Document, List[str]):
    """
    Read a Humdrum **kern file.

    Args:
        path (Union[str, Path]): File path to read
        strict (Optional[bool]): If True, raise an error if the **kern file has any errors. Otherwise, return a list of errors.

    Returns (Document, List[str]): Document object and list of error messages. Empty list if no errors.

    Examples:
        >>> import kernpy as kp
        >>> document, _ = kp.read('path/to/file.krn')

        >>> document, errors = kp.read('path/to/file.krn')
        >>> if len(errors) > 0:
        >>>     print(errors)
        ['Error: Invalid **kern spine: 1', 'Error: Invalid **kern spine: 2']
    """
    return Generic.read(
        path=Path(path),
        strict=strict
    )


@deprecated("Use 'loads' instead.")
def create(
        content: str,
        strict=False
) -> (Document, []):
    """
    Create a Document object from a string encoded in Humdrum **kern format.

    Args:
        content: String encoded in Humdrum **kern format
        strict: If True, raise an error if the **kern file has any errors. Otherwise, return a list of errors.

    Returns (Document, list): Document object and list of error messages. Empty list if no errors.

    Examples:
        >>> import kernpy as kp
        >>> document, errors = kp.create('**kern\n4e\n4f\n4g\n*-\n')
        >>> if len(errors) > 0:
        >>>     print(errors)
        ['Error: Invalid **kern spine: 1', 'Error: Invalid **kern spine: 2']
    """
    return Generic.create(
        content=content,
        strict=strict
    )


@deprecated("Use 'dumps' instead.")
def export(
        document: Document,
        options: ExportOptions
) -> str:
    """
    Export a Document object to a string.

    Args:
        document: Document object to export
        options: Export options

    Returns: Exported string

    Examples:
        >>> import kernpy as kp
        >>> document, errors = kp.read('path/to/file.krn')
        >>> options = kp.ExportOptions()
        >>> content = kp.export(document, options)
    """
    return Generic.export(
        document=document,
        options=options
    )


@deprecated("Use 'dump' instead.")
def store(
        document: Document,
        path: Union[str, Path],
        options: ExportOptions
) -> None:
    """
    Store a Document object to a file.

    Args:
        document (Document): Document object to store
        path (Union[str, Path]): File path to store
        options (ExportOptions): Export options

    Returns: None

    Examples:
        >>> import kernpy as kp
        >>> document, errors = kp.read('path/to/file.krn')
        >>> options = kp.ExportOptions()
        >>> kp.store(document, 'path/to/store.krn', options)

    """
    Generic.store(
        document=document,
        path=Path(path),
        options=options
    )


@deprecated("Use 'graph' instead.")
def store_graph(
        document: Document,
        path: Union[str, Path]
) -> None:
    """
    Create a graph representation of a Document object using Graphviz. Save the graph to a file.

    Args:
        document (Document): Document object to create graph from
        path (str): File path to save the graph

    Returns (None): None

    Examples:
        >>> import kernpy as kp
        >>> document, errors = kp.read('path/to/file.krn')
        >>> kp.store_graph(document, 'path/to/graph.dot')
    """
    return Generic.store_graph(
        document=document,
        path=Path(path)
    )


@deprecated("Use 'spine_types' instead.")
def get_spine_types(
        document: Document,
        spine_types: Optional[Sequence[str]] = None
) -> List[str]:
    """
    Get the spines of a Document object.

    Args:
        document (Document): Document object to get spines from
        spine_types (Optional[Sequence[str]]): List of spine types to get. If None, all spines are returned.

    Returns (List[str]): List of spines

    Examples:
        >>> import kernpy as kp
        >>> document, _ = kp.read('path/to/file.krn')
        >>> kp.get_spine_types(document)
        ['**kern', '**kern', '**kern', '**kern', '**root', '**harm']
        >>> kp.get_spine_types(document, None)
        ['**kern', '**kern', '**kern', '**kern', '**root', '**harm']
        >>> kp.get_spine_types(document, ['**kern'])
        ['**kern', '**kern', '**kern', '**kern']
        >>> kp.get_spine_types(document, ['**kern', '**root'])
        ['**kern', '**kern', '**kern', '**kern', '**root']
        >>> kp.get_spine_types(document, ['**kern', '**root', '**harm'])
        ['**kern', '**kern', '**kern', '**kern', '**root', '**harm']
        >>> kp.get_spine_types(document, [])
        []
    """
    return Generic.get_spine_types(
        document=document,
        spine_types=spine_types
    )



from __future__ import annotations

import sys
from pathlib import Path
from unicodedata import category

from kernpy.core import Token, SpineOperationToken
from kernpy.core.document import MultistageTree, Node


class GraphvizExporter:
    def export_token(self, token: Token):
        if token is None or token.encoding is None:
            return ''
        else:
            return token.encoding.replace('\"', '\\"').replace('\\', '\\\\')

    @staticmethod
    def node_id(node: Node):
        return f"node{id(node)}"

    def export_to_dot(self, tree: MultistageTree, filename: Path = None):
        """
        Export the given MultistageTree to DOT format.

        Args:
            tree (MultistageTree): The tree to export.
            filename (Path or None): The output file path. If None, prints to stdout.
        """
        file = sys.stdout if filename is None else open(filename, 'w')

        try:
            file.write('digraph G {\n')
            file.write('    node [shape=record];\n')
            file.write('    rankdir=TB;\n')  # Ensure top-to-bottom layout

            # Create subgraphs for each stage
            for stage_index, stage in enumerate(tree.stages):
                if stage:
                    file.write('  {rank=same; ')
                    for node in stage:
                        file.write(f'"{self.node_id(node)}"; ')
                    file.write('}\n')

            # Write nodes and their connections
            self._write_nodes_iterative(tree.root, file)
            self._write_edges_iterative(tree.root, file)

            file.write('}\n')

        finally:
            if filename is not None:
                file.close()  # Close only if we explicitly opened a file

    def _write_nodes_iterative(self, root, file):
        stack = [root]

        while stack:
            node = stack.pop()
            header_label = f'header #{node.header_node.id}' if node.header_node else ''
            last_spine_operator_label = f'last spine op. #{node.last_spine_operator_node.id}' if node.last_spine_operator_node else ''
            category_name = getattr(getattr(getattr(node, "token", None), "category", None), "_name_", "Non defined category")


            top_record_label = f'{{ #{node.id}| stage {node.stage} | {header_label} | {last_spine_operator_label} | {category_name} }}'
            signatures_label = ''
            if node.last_signature_nodes and node.last_signature_nodes.nodes:
                for k, v in node.last_signature_nodes.nodes.items():
                    if signatures_label:
                        signatures_label += '|'
                    signatures_label += f'{k} #{v.id}'

            if isinstance(node.token, SpineOperationToken) and node.token.cancelled_at_stage:
                signatures_label += f'| {{ cancelled at stage {node.token.cancelled_at_stage} }}'

            file.write(f'  "{self.node_id(node)}" [label="{{ {top_record_label} | {signatures_label} | {self.export_token(node.token)} }}"];\n')

            # Add children to the stack to be processed
            for child in reversed(node.children):
                stack.append(child)

    def _write_edges_iterative(self, root, file):
        stack = [root]

        while stack:
            node = stack.pop()
            for child in node.children:
                file.write(f'  "{self.node_id(node)}" -> "{self.node_id(child)}";\n')
                stack.append(child)



from __future__ import annotations
from typing import Optional

from .base_antlr_spine_parser_listener import BaseANTLRSpineParserListener
from .spine_importer import SpineImporter
from .tokens import SimpleToken, TokenCategory, Token
from .kern_spine_importer import KernSpineImporter, KernSpineListener
from .dynam_spine_importer import DynamSpineImporter


class DynSpineImporter(SpineImporter):
    def __init__(self, verbose: Optional[bool] = False):
        """
        KernSpineImporter constructor.

        Args:
            verbose (Optional[bool]): Level of verbosity for error messages.
        """
        super().__init__(verbose=verbose)

    def import_listener(self) -> BaseANTLRSpineParserListener:
        return KernSpineListener()

    def import_token(self, encoding: str) -> Token:
        # TODO: Find out differences between **dyn vs **dynam and change this class. Using the same dor both for now.
        dynam_importer = DynamSpineImporter()
        return dynam_importer.import_token(encoding)






from __future__ import annotations
from typing import Optional

from .base_antlr_spine_parser_listener import BaseANTLRSpineParserListener
from .spine_importer import SpineImporter
from .tokens import Token


class MensSpineImporter(SpineImporter):
    def __init__(self, verbose: Optional[bool] = False):
        """
        MensSpineImporter constructor.

        Args:
            verbose (Optional[bool]): Level of verbosity for error messages.
        """
        super().__init__(verbose=verbose)

    def import_listener(self) -> BaseANTLRSpineParserListener:
        raise NotImplementedError()

    def import_token(self, encoding: str) -> Token:
        raise NotImplementedError()



from __future__ import annotations

from copy import deepcopy
from typing import Optional

from .pitch_models import (
    NotationEncoding,
    AgnosticPitch,
    PitchExporterFactory,
    PitchImporterFactory,
    Direction,
)


Intervals = {
    -2: 'dd1',
    -1: 'd1',
    0: 'P1',
    1: 'A1',
    2: 'AA1',
    3: 'dd2',
    4: 'd2',
    5: 'm2',
    6: 'M2',
    7: 'A2',
    8: 'AA2',
    9: 'dd3',
    10: 'd3',
    11: 'm3',
    12: 'M3',
    13: 'A3',
    14: 'AA3',
    15: 'dd4',
    16: 'd4',
    17: 'P4',
    18: 'A4',
    19: 'AA4',
    21: 'dd5',
    # 20 is unused
    22: 'd5',
    23: 'P5',
    24: 'A5',
    25: 'AA5',
    26: 'dd6',
    27: 'd6',
    28: 'm6',
    29: 'M6',
    30: 'A6',
    31: 'AA6',
    32: 'dd7',
    33: 'd7',
    34: 'm7',
    35: 'M7',
    36: 'A7',
    37: 'AA7',
    40: 'octave'
}
"""
Base-40 interval classes (d=diminished, m=minor, M=major, P=perfect, A=augmented)
"""

IntervalsByName = {v: k for k, v in Intervals.items()}  # reverse the key-value pairs
AVAILABLE_INTERVALS = sorted(IntervalsByName.keys())

LETTER_TO_SEMITONES = {
    'C': 0,
    'D': 2,
    'E': 4,
    'F': 5,
    'G': 7,
    'A': 9,
    'B': 11,
}


def transpose_agnostics(
        input_pitch: AgnosticPitch,
        interval: int,
        direction: str = Direction.UP.value
) -> AgnosticPitch:
    """
    Transpose an AgnosticPitch by a given interval.

    Args:
        input_pitch (AgnosticPitch): The pitch to transpose.
        interval (int): The interval to transpose the pitch.
        direction (str): The direction of the transposition. 'UP' or 'DOWN'. Default is 'UP'.

    Returns :
        AgnosticPitch: The transposed pitch.

    Examples:
        >>> transpose_agnostics(AgnosticPitch('C', 4), IntervalsByName['P4'])
        AgnosticPitch('F', 4)
        >>> transpose_agnostics(AgnosticPitch('C', 4), IntervalsByName['P4'], direction='down')
        AgnosticPitch('G', 3)
        >>> transpose_agnostics(AgnosticPitch('C#', 4), IntervalsByName['P4'])
        AgnosticPitch('F#', 4)
        >>> transpose_agnostics(AgnosticPitch('G', 4), IntervalsByName['m3'], direction='down')
        AgnosticPitch('Bb', 4)

    """
    return AgnosticPitch.to_transposed(input_pitch, interval, direction)


def transpose_encoding_to_agnostic(
        input_encoding: str,
        interval: int,
        input_format: str = NotationEncoding.HUMDRUM.value,
        direction: str = Direction.UP.value
) -> AgnosticPitch:
    """
    Transpose a pitch by a given interval.

    The pitch must be in the American notation.

    Args:
        input_encoding (str): The pitch to transpose.
        interval (int): The interval to transpose the pitch.
        input_format (str): The encoding format of the pitch. Default is HUMDRUM.
        direction (str): The direction of the transposition.'UP' or 'DOWN' Default is 'UP'.

    Returns:
        AgnosticPitch: The transposed pitch.

    Examples:
        >>> transpose_encoding_to_agnostic('ccc', IntervalsByName['P4'], input_format='kern')
        AgnosticPitch('fff', 4)
        >>> transpose_encoding_to_agnostic('ccc', IntervalsByName['P4'], input_format=NotationEncoding.HUMDRUM.value)
        AgnosticPitch('fff', 4)
        >>> transpose_encoding_to_agnostic('ccc', IntervalsByName['P4'], input_format='kern', direction='down')
        AgnosticPitch('gg', 3)
        >>> transpose_encoding_to_agnostic('ccc#', IntervalsByName['P4'])
        AgnosticPitch('fff#', 4)
        >>> transpose_encoding_to_agnostic('G4', IntervalsByName['m3'], input_format='american')
        AgnosticPitch('Bb4', 4)
        >>> transpose_encoding_to_agnostic('C3', IntervalsByName['P4'], input_format='american', direction='down')
        AgnosticPitch('G2', 2)

    """
    importer = PitchImporterFactory.create(input_format)
    pitch: AgnosticPitch = importer.import_pitch(input_encoding)

    return transpose_agnostics(pitch, interval, direction=direction)


def transpose_agnostic_to_encoding(
        agnostic_pitch: AgnosticPitch,
        interval: int,
        output_format: str = NotationEncoding.HUMDRUM.value,
        direction: str = Direction.UP.value
) -> str:
    """
    Transpose an AgnosticPitch by a given interval.

    Args:
        agnostic_pitch (AgnosticPitch): The pitch to transpose.
        interval (int): The interval to transpose the pitch.
        output_format (Optional[str]): The encoding format of the transposed pitch. Default is HUMDRUM.
        direction (Optional[str]): The direction of the transposition.'UP' or 'DOWN' Default is 'UP'.

    Returns (str):
        str: The transposed pitch.

    Examples:
        >>> transpose_agnostic_to_encoding(AgnosticPitch('C', 4), IntervalsByName['P4'])
        'F4'
        >>> transpose_agnostic_to_encoding(AgnosticPitch('C', 4), IntervalsByName['P4'], direction='down')
        'G3'
        >>> transpose_agnostic_to_encoding(AgnosticPitch('C#', 4), IntervalsByName['P4'])
        'F#4'
        >>> transpose_agnostic_to_encoding(AgnosticPitch('G', 4), IntervalsByName['m3'], direction='down')
        'Bb4'
    """
    exporter = PitchExporterFactory.create(output_format)
    transposed_pitch = transpose_agnostics(agnostic_pitch, interval, direction=direction)
    content = exporter.export_pitch(transposed_pitch)

    return content





def transpose(
        input_encoding: str,
        interval: int,
        input_format: str = NotationEncoding.HUMDRUM.value,
        output_format: str = NotationEncoding.HUMDRUM.value,
        direction: str = Direction.UP.value
) -> str:
    """
    Transpose a pitch by a given interval.

    The pitch must be in the American notation.

    Args:
        input_encoding (str): The pitch to transpose.
        interval (int): The interval to transpose the pitch.
        input_format (str): The encoding format of the pitch. Default is HUMDRUM.
        output_format (str): The encoding format of the transposed pitch. Default is HUMDRUM.
        direction (str): The direction of the transposition.'UP' or 'DOWN' Default is 'UP'.

    Returns:
        str: The transposed pitch.

    Examples:
        >>> transpose('ccc', IntervalsByName['P4'], input_format='kern', output_format='kern')
        'fff'
        >>> transpose('ccc', IntervalsByName['P4'], input_format=NotationEncoding.HUMDRUM.value)
        'fff'
        >>> transpose('ccc', IntervalsByName['P4'], input_format='kern', direction='down')
        'gg'
        >>> transpose('ccc', IntervalsByName['P4'], input_format='kern', direction=Direction.DOWN.value)
        'gg'
        >>> transpose('ccc#', IntervalsByName['P4'])
        'fff#'
        >>> transpose('G4', IntervalsByName['m3'], input_format='american')
        'Bb4'
        >>> transpose('G4', IntervalsByName['m3'], input_format=NotationEncoding.AMERICAN.value)
        'Bb4'
        >>> transpose('C3', IntervalsByName['P4'], input_format='american', direction='down')
        'G2'


    """
    importer = PitchImporterFactory.create(input_format)
    pitch: AgnosticPitch = importer.import_pitch(input_encoding)

    transposed_pitch = transpose_agnostics(pitch, interval, direction=direction)

    exporter = PitchExporterFactory.create(output_format)
    content = exporter.export_pitch(transposed_pitch)

    return content


def agnostic_distance(
    first_pitch: AgnosticPitch,
    second_pitch: AgnosticPitch,
) -> int:
    """
    Calculate the distance in semitones between two pitches.

    Args:
        first_pitch (AgnosticPitch): The first pitch to compare.
        second_pitch (AgnosticPitch): The second pitch to compare.

    Returns:
        int: The distance in semitones between the two pitches.

    Examples:
        >>> agnostic_distance(AgnosticPitch('C4'), AgnosticPitch('E4'))
        4
        >>> agnostic_distance(AgnosticPitch('C4'), AgnosticPitch('B3'))
        -1
    """
    def semitone_index(p: AgnosticPitch) -> int:
        # base letter:
        letter = p.name.replace('+', '').replace('-', '')
        base = LETTER_TO_SEMITONES[letter]
        # accidentals: '+' is one sharp, '-' one flat
        alteration = p.name.count('+') - p.name.count('-')
        return p.octave * 12 + base + alteration

    return semitone_index(second_pitch) - semitone_index(first_pitch)


def distance(
    first_encoding: str,
    second_encoding: str,
    *,
    first_format: str = NotationEncoding.HUMDRUM.value,
    second_format: str = NotationEncoding.HUMDRUM.value,
) -> int:
    """
    Calculate the distance in semitones between two pitches.

    Args:
        first_encoding (str): The first pitch to compare.
        second_encoding (str): The second pitch to compare.
        first_format (str): The encoding format of the first pitch. Default is HUMDRUM.
        second_format (str): The encoding format of the second pitch. Default is HUMDRUM.

    Returns:
        int: The distance in semitones between the two pitches.

    Examples:
        >>> distance('C4', 'E4')
        4
        >>> distance('C4', 'B3')
        -1
    """
    first_importer = PitchImporterFactory.create(first_format)
    first_pitch: AgnosticPitch = first_importer.import_pitch(first_encoding)

    second_importer = PitchImporterFactory.create(second_format)
    second_pitch: AgnosticPitch = second_importer.import_pitch(second_encoding)

    return agnostic_distance(first_pitch, second_pitch)



from .generated.kernSpineParserListener import kernSpineParserListener
from .generated.kernSpineParser import kernSpineParser
from .tokens import BarToken, SimpleToken, TokenCategory, Subtoken, ChordToken, BoundingBox, \
    BoundingBoxToken, ClefToken, KeySignatureToken, TimeSignatureToken, MeterSymbolToken, BarToken, NoteRestToken, \
    KeyToken, InstrumentToken

from typing import List


class BaseANTLRSpineParserListener(kernSpineParserListener):
    def __init__(self):
        self.token = None

        self.first_chord_element = None
        self.chord_tokens = None
        self.duration_subtokens = []
        self.diatonic_pitch_and_octave_subtoken = None
        self.accidental_subtoken = None
        # self.decorations = {}  # in order to standardize the order of decorators, we map the different properties to their class names
        # We cannot order it using the class name because there are rules with subrules, such as ties, or articulations. We order it using the encoding itself
        self.decorations: List[Subtoken] = []
        self.in_chord = False
        # self.page_start_rows = [] # TODO
        self.measure_start_rows = []
        self.last_bounding_box = None

    def _add_decoration(self, new_decoration: Subtoken):
        """
        Private method for ensuring there are no duplicated decorations.

        It's like a set sorted data structure.

        Args:
            new_decoration (Subtoken): The new decoration to add.

        Returns (None):
            None
        """
        for existing_decoration in self.decorations:
            if existing_decoration.encoding == new_decoration.encoding:
                return
        self.decorations.append(new_decoration)

    def enterStart(self, ctx: kernSpineParser.StartContext):
        self.token = None
        self.duration_subtokens = []
        self.diatonic_pitch_and_octave_subtoken = None
        self.accidental_subtoken = None
        # self.decorations = {}
        self.decorations = []

    # def process_decorations(self, ctx: ParserRuleContext):
    #     # in order to standardize the order of note decorators, we map the different properties to their class names
    #     decorations = {}
    #
    #     for child in ctx.getChildren():
    #         # all decorations have just a child
    #         if child.getChildCount() != 1:
    #             raise Exception('Only 1 decoration child expected, and found ' + child.getChildCount() + ', check '
    #                                                                                                      'the '
    #                                                                                                      'grammar')
    #         clazz = type(child.getChild(0))
    #         decoration_type = clazz.__name__
    #         if decoration_type in decorations:
    #             logging.warning(
    #                 f'The decoration {decoration_type} is duplicated')  # TODO Dar información de línea, columna - ¿lanzamos excepción? - hay algunas que sí pueden estar duplicadas? Barrados?
    #         decorations[decoration_type] = child.getText()
    #     for key in sorted(decorations.keys()):
    #         subtoken = Subtoken(decorations[key], TokenCategory.DECORATION)
    #         self.duration_subtoken.append(subtoken)

    def exitDuration(self, ctx: kernSpineParser.DurationContext):
        self.duration_subtokens = [Subtoken(ctx.modernDuration().getText(), TokenCategory.DURATION)]
        for i in range(len(ctx.augmentationDot())):
            self.duration_subtokens.append(Subtoken(".", TokenCategory.DURATION))  # TODO: Add a new TokenCategory for dots, inherit from DURATION.

        if ctx.graceNote():
            self.duration_subtokens.append(Subtoken(ctx.graceNote().getText(), TokenCategory.DURATION))  # TODO: Add a new TokenCategory for grace notes, inherit from DURATION.

        if ctx.appoggiatura():
            self.duration_subtokens.append(Subtoken(ctx.appoggiatura().getText(), TokenCategory.DURATION))  # TODO: Add a new TokenCategory for appoggiaturas, inherit from DURATION.

    def exitDiatonicPitchAndOctave(self, ctx: kernSpineParser.DiatonicPitchAndOctaveContext):
        self.diatonic_pitch_and_octave_subtoken = Subtoken(ctx.getText(), TokenCategory.PITCH)

    def exitNoteDecoration(self, ctx: kernSpineParser.NoteDecorationContext):
        # clazz = type(ctx.getChild(0))
        # decoration_type = clazz.__name__
        # if decoration_type in self.decorations:
        #    logging.warning(
        #        f'The decoration {decoration_type} is duplicated after reading {ctx.getText()}')  # TODO Dar información de línea, columna - ¿lanzamos excepción? - hay algunas que sí pueden estar duplicadas? Barrados?

        # self.decorations[decoration_type] = ctx.getText()
        # We cannot order it using the class name because there are rules with subrules, such as ties, or articulations. We order it using the encoding itself. NOT YET!
        decoration_encoding = ctx.getText()
        decoration_subtoken = Subtoken(decoration_encoding, TokenCategory.DECORATION)
        self._add_decoration(decoration_subtoken)

    def exitRestDecoration(self, ctx: kernSpineParser.NoteDecorationContext):
        # clazz = type(ctx.getChild(0))
        # decoration_type = clazz.__name__
        # if decoration_type in self.decorations:
        #    logging.warning(
        #        f'The decoration {decoration_type} is duplicated after reading {ctx.getText()}')  # TODO Dar información de línea, columna - ¿lanzamos excepción? - hay algunas que sí pueden estar duplicadas? Barrados?

        # self.decorations[decoration_type] = ctx.getText()
        # We cannot order it using the class name because there are rules with subrules, such as ties, or articulations. We order it using the encoding itself
        decoration = ctx.getText()
        if decoration != '/' and decoration != '\\':
            decoration_encoding = ctx.getText()
            decoration_subtoken = Subtoken(decoration_encoding, TokenCategory.DECORATION)
            self._add_decoration(decoration_subtoken)

    def addNoteRest(self, ctx, pitchduration_subtokens):
        # subtoken = Subtoken(self.decorations[key], TokenCategory.DECORATION)
        token = NoteRestToken(ctx.getText(), pitchduration_subtokens, self.decorations)
        if self.in_chord:
            self.chord_tokens.append(token)
        else:
            self.token = token

    def exitNote(self, ctx: kernSpineParser.NoteContext):
        pitch_duration_tokens = []
        for duration_subtoken in self.duration_subtokens:
            pitch_duration_tokens.append(duration_subtoken)
        pitch_duration_tokens.append(self.diatonic_pitch_and_octave_subtoken)
        if ctx.alteration():
            pitch_duration_tokens.append(Subtoken(ctx.alteration().getText(), TokenCategory.ALTERATION))

        self.addNoteRest(ctx, pitch_duration_tokens)

    def exitRest(self, ctx: kernSpineParser.RestContext):
        pitch_duration_tokens = []
        for duration_subtoken in self.duration_subtokens:
            pitch_duration_tokens.append(duration_subtoken)
        pitch_duration_tokens.append(Subtoken('r', TokenCategory.REST))
        self.addNoteRest(ctx, pitch_duration_tokens)

    def enterChord(self, ctx: kernSpineParser.ChordContext):
        self.in_chord = True
        self.chord_tokens = []

    def exitChord(self, ctx: kernSpineParser.ChordContext):
        self.in_chord = False
        self.token = ChordToken(ctx.getText(), TokenCategory.CHORD, self.chord_tokens)

    def exitEmpty(self, ctx: kernSpineParser.EmptyContext):
        self.token = SimpleToken(ctx.getText(), TokenCategory.EMPTY)

    def exitNonVisualTandemInterpretation(self, ctx: kernSpineParser.NonVisualTandemInterpretationContext):
        self.token = SimpleToken(ctx.getText(), TokenCategory.OTHER)

    def exitVisualTandemInterpretation(self, ctx: kernSpineParser.VisualTandemInterpretationContext):
        self.token = SimpleToken(ctx.getText(), TokenCategory.ENGRAVED_SYMBOLS)

    def exitOtherContextual(self, ctx: kernSpineParser.ContextualContext):
        self.token = SimpleToken(ctx.getText(), TokenCategory.OTHER_CONTEXTUAL)

    def exitClef(self, ctx: kernSpineParser.ClefContext):
        self.token = ClefToken(ctx.getText())

    def exitKeySignature(self, ctx: kernSpineParser.KeySignatureContext):
        self.token = KeySignatureToken(ctx.getText())

    def exitKeyCancel(self, ctx: kernSpineParser.KeyCancelContext):
        self.token = KeySignatureToken(ctx.getText())

    def exitKey(self, ctx: kernSpineParser.KeyContext):
        self.token = KeyToken(ctx.getText())

    def exitTimeSignature(self, ctx: kernSpineParser.TimeSignatureContext):
        self.token = TimeSignatureToken(ctx.getText())

    def exitMeterSymbol(self, ctx: kernSpineParser.MeterSymbolContext):
        self.token = MeterSymbolToken(ctx.getText())

    def exitStructural(self, ctx: kernSpineParser.StructuralContext):
        self.token = SimpleToken(ctx.getText(), TokenCategory.STRUCTURAL)

    def exitXywh(self, ctx: kernSpineParser.XywhContext):
        self.last_bounding_box = BoundingBox(int(ctx.x().getText()), int(ctx.y().getText()), int(ctx.w().getText()),
                                             int(ctx.h().getText()))

    def exitBoundingBox(self, ctx: kernSpineParser.BoundingBoxContext):
        page = ctx.pageNumber().getText()
        bbox = BoundingBox(int(ctx.xywh().x().getText()), int(ctx.xywh().y().getText()), int(ctx.xywh().w().getText()),
                           int(ctx.xywh().h().getText()))
        self.token = BoundingBoxToken(ctx.getText(), page, bbox)

    def exitInstrument(self, ctx: kernSpineParser.InstrumentContext):
        self.token = InstrumentToken(ctx.getText())


    def exitBarline(self, ctx: kernSpineParser.BarlineContext):
        txt_without_number = ''
        if ctx.EQUAL(0) and ctx.EQUAL(1):
            txt_without_number = '=='
        elif ctx.EQUAL(0):
            txt_without_number = '='
        if ctx.barLineType():
            txt_without_number += ctx.barLineType().getText()
        if ctx.fermata():
            txt_without_number += ctx.fermata().getText()

        # correct wrong encodings
        if txt_without_number == ':!:':
            txt_without_number = ':|!|:'
        elif txt_without_number == ':|!|:':
            txt_without_number = ':|!|:'

        self.token = BarToken(txt_without_number)
        self.token.hidden = "-" in ctx.getText()  # hidden




from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Optional

from antlr4 import InputStream, CommonTokenStream, ParseTreeWalker, BailErrorStrategy, \
    PredictionMode

from .generated.kernSpineLexer import kernSpineLexer
from .generated.kernSpineParser import kernSpineParser
from .base_antlr_spine_parser_listener import BaseANTLRSpineParserListener
from .error_listener import ErrorListener
from .tokens import Token


class SpineImporter(ABC):
    def __init__(self, verbose: Optional[bool] = False):
        """
        SpineImporter constructor.
        This class is an abstract base class for importing all kinds of spines.

        Args:
            verbose (Optional[bool]): Level of verbosity for error messages.
        """
        self.import_listener = self.import_listener()
        self.error_listener = ErrorListener(verbose=verbose)

    @abstractmethod
    def import_listener(self) -> BaseANTLRSpineParserListener:
        pass

    @abstractmethod
    def import_token(self, encoding: str) -> Token:
        pass

    @classmethod
    def _raise_error_if_wrong_input(cls, encoding: str):
        if encoding is None:
            raise ValueError("Encoding cannot be None")
        if not isinstance(encoding, str):
            raise TypeError("Encoding must be a string")
        if encoding == '':
            raise ValueError("Encoding cannot be an empty string")





from __future__ import annotations
from typing import Optional

from antlr4 import InputStream, CommonTokenStream, ParseTreeWalker, BailErrorStrategy, \
    PredictionMode

from .base_antlr_importer import BaseANTLRListenerImporter
from .base_antlr_spine_parser_listener import BaseANTLRSpineParserListener
from .error_listener import ErrorListener
from .kern_spine_importer import KernSpineListener, KernSpineImporter
from .spine_importer import SpineImporter
from .generated.kernSpineLexer import kernSpineLexer
from .generated.kernSpineParser import kernSpineParser
from .tokens import TokenCategory, Token, SimpleToken


class RootSpineListener(BaseANTLRSpineParserListener):
    def __init__(self):
        super().__init__()


class RootListenerImporter(BaseANTLRListenerImporter):

    def createListener(self):
        return KernSpineListener()

    def createLexer(self, tokenStream):
        return kernSpineLexer(tokenStream)

    def createParser(self, tokenStream):
        return kernSpineParser(tokenStream)

    def startRule(self):
        return self.parser.start()


class RootSpineImporter(SpineImporter):
    def __init__(self, verbose: Optional[bool] = False):
        """
        KernSpineImporter constructor.

        Args:
            verbose (Optional[bool]): Level of verbosity for error messages.
        """
        super().__init__(verbose=verbose)

    def import_listener(self) -> BaseANTLRSpineParserListener:
        #return RootSpineListener() # TODO: Create a custom functional listener for RootSpineImporter
        return KernSpineListener()

    def import_token(self, encoding: str) -> Token:
        self._raise_error_if_wrong_input(encoding)

        kern_spine_importer = KernSpineImporter()
        token = kern_spine_importer.import_token(encoding)

        return token  # The **root spine tokens are always a subset of the **kern spine tokens




import warnings
import functools


def deprecated(reason: str):
    """
    Decorator to mark a function or class as deprecated.
    
    Args:
        reason (str): The reason why the function/class is deprecated.
    
    Example:
        @deprecated("Use new_function instead.")
        def old_function():
            pass
    """
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            warnings.warn(
                f"'{func.__name__}' is deprecated: {reason}",
                category=DeprecationWarning,
                stacklevel=2,
            )
            return func(*args, **kwargs)
        return wrapper
    return decorator


def deprecated_argument(arg_name: str, reason: str):
    """
    Decorator to mark a specific argument as deprecated.
    
    Args:
        arg_name (str): The deprecated argument name.
        reason (str): The reason why the argument is deprecated.
    
    Example:
        @deprecated_argument("old_param", "Use 'new_param' instead.")
        def my_function(new_param, old_param=None):
            pass
    """
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if arg_name in kwargs:
                warnings.warn(
                    f"Argument '{arg_name}' in '{func.__name__}' is deprecated: {reason}",
                    category=DeprecationWarning,
                    stacklevel=2,
                )
            return func(*args, **kwargs)
        return wrapper
    return decorator




# Run from the root project (the 'test' parent folder): python3 -m unittest test/test_importer.py
# or from the IDE
import os
import unittest
import logging
import sys
import tempfile
from PIL import Image


from kernpy.polish_scores.download_polish_dataset import convert_and_download_file

logger = logging.getLogger()
logger.level = logging.INFO  # change it DEBUG to trace errors
logger.addHandler(logging.StreamHandler(sys.stdout))


class DownloadPolishScoresTestCase(unittest.TestCase):
    def check_image_sizes_equal(self, image_path1, image_path2):
        try:
            # Open the images
            image1 = Image.open(image_path1)
            image2 = Image.open(image_path2)

            # Get the sizes of the images
            size1 = image1.size
            size2 = image2.size

            # Check if the sizes are equal
            self.assertEqual(size1, size2)
        except IOError:
            raise Exception("Unable to load one or both images")

    def checkEqualFiles(self, file1, file2):
        with open(file1, 'r') as file1, open(file2, 'r') as file2:
            content1 = file1.read()
            content2 = file2.read()
            self.assertEqual(content1, content2)

    @unittest.skip("This test is too slow. Maybe we should mock the download. Comment this line to run the test")
    def test_convert_and_download_file(self):
        temp_dir = tempfile.mkdtemp()
        input_folder = 'resource_dir/polish/test1'
        log_file = os.path.join(temp_dir, 'polish_index.json')
        logging.info(f'Writing DownloadPolishScoresTestCase test to folder {temp_dir}')
        convert_and_download_file(input_folder + '/pl-wn--mus-iii-118-771--003_badarzewska-tekla--mazurka-brillante.krn', temp_dir, log_file)
        for i in range(9, 11):
            self.checkEqualFiles(input_folder + f'/pages/{i}.ekrn', temp_dir + f'/{i}.ekrn')
            self.check_image_sizes_equal(input_folder + f'/pages/{i}.jpg', temp_dir + f'/{i}.jpg')

    def test_polish_package_import(self):
        from kernpy.polish_scores import download_polish_scores
        self.assertIsNotNone(download_polish_scores)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from typing import Optional

from .kern_spine_importer import KernSpineListener, KernSpineImporter
from .base_antlr_spine_parser_listener import BaseANTLRSpineParserListener
from .spine_importer import SpineImporter
from .tokens import MHXMToken, Token, TokenCategory, SimpleToken


class MxhmSpineImporter(SpineImporter):
    def __init__(self, verbose: Optional[bool] = False):
        """
        KernSpineImporter constructor.

        Args:
            verbose (Optional[bool]): Level of verbosity for error messages.
        """
        super().__init__(verbose=verbose)

    def import_listener(self) -> BaseANTLRSpineParserListener:
        return KernSpineListener()

    def import_token(self, encoding: str) -> Token:
        self._raise_error_if_wrong_input(encoding)

        try:
            kern_spine_importer = KernSpineImporter()
            token = kern_spine_importer.import_token(encoding)
        except Exception as e:
            return SimpleToken(encoding, TokenCategory.HARMONY)

        ACCEPTED_CATEGORIES = {
            TokenCategory.STRUCTURAL,
            TokenCategory.SIGNATURES,
            TokenCategory.EMPTY,
            TokenCategory.IMAGE_ANNOTATIONS,
            TokenCategory.BARLINES,
            TokenCategory.COMMENTS,
        }

        if any(TokenCategory.is_child(child=token.category, parent=cat) for cat in ACCEPTED_CATEGORIES):
            return SimpleToken(encoding, TokenCategory.HARMONY)

        return token



import os
import unittest
import sys
from unittest.mock import patch

import kernpy as kp


class TokenCategoryHierarchyTestCase(unittest.TestCase):
    def _is_child(cls, parent, child):
        cls.assertTrue(kp.TokenCategoryHierarchyMapper.is_child(parent=parent, child=child))

    def _is_not_child(cls, parent, child):
        cls.assertFalse(kp.TokenCategoryHierarchyMapper.is_child(parent=parent, child=child))

    def _is_valid_category(cls, include, exclude, expected_categories):
        cls.assertSetEqual(
            expected_categories,
            kp.TokenCategoryHierarchyMapper.valid(include=include, exclude=exclude)
        )

    def test_is_child_of_CORE(self):
        self._is_child(kp.TokenCategory.CORE, kp.TokenCategory.NOTE_REST)
        self._is_child(kp.TokenCategory.CORE, kp.TokenCategory.NOTE)
        self._is_child(kp.TokenCategory.CORE, kp.TokenCategory.REST)
        self._is_child(kp.TokenCategory.CORE, kp.TokenCategory.DURATION)
        self._is_child(kp.TokenCategory.CORE, kp.TokenCategory.CHORD)
        self._is_child(kp.TokenCategory.CORE, kp.TokenCategory.EMPTY)

    def test_is_child_of_SIGNATURES(self):
        self._is_child(kp.TokenCategory.SIGNATURES, kp.TokenCategory.CLEF)
        self._is_child(kp.TokenCategory.SIGNATURES, kp.TokenCategory.TIME_SIGNATURE)
        self._is_child(kp.TokenCategory.SIGNATURES, kp.TokenCategory.METER_SYMBOL)
        self._is_child(kp.TokenCategory.SIGNATURES, kp.TokenCategory.KEY_SIGNATURE)

    def test_not_child_across_branches(self):
        """
        Test that categories from different branches are not related.
        """
        # For example, CLEF (under SIGNATURES) is not a child of CORE.
        self._is_not_child(kp.TokenCategory.CORE, kp.TokenCategory.CLEF)
        # Similarly, dynamics is not a child of comments.
        self._is_not_child(kp.TokenCategory.COMMENTS, kp.TokenCategory.DYNAMICS)

    def test_invalid_parent_type(self):
        self.assertFalse(kp.TokenCategoryHierarchyMapper.is_child("CORE", kp.TokenCategory.NOTE))

    def test_invalid_child_type(self):
        self.assertFalse(kp.TokenCategoryHierarchyMapper.is_child(kp.TokenCategory.CORE, "NOTE"))

    def test_child_of_same_category(self):
        self.assertTrue(kp.TokenCategoryHierarchyMapper.is_child(kp.TokenCategory.CORE, kp.TokenCategory.CORE))
        self.assertTrue(kp.TokenCategoryHierarchyMapper.is_child(kp.TokenCategory.HARMONY, kp.TokenCategory.HARMONY))


    def test_match_included_category(self):
        """Test that a category matches when it's explicitly included."""
        self.assertTrue(
            kp.TokenCategoryHierarchyMapper.match(kp.TokenCategory.NOTE, include={kp.TokenCategory.NOTE_REST})
        )

    def test_match_with_exclusion(self):
        """Test that a category is correctly excluded."""
        self.assertFalse(
            kp.TokenCategoryHierarchyMapper.match(kp.TokenCategory.NOTE, include={kp.TokenCategory.NOTE_REST}, exclude={kp.TokenCategory.NOTE})
        )

    def test_match_without_includes(self):
        """Test that a category is included when include is None."""
        self.assertTrue(
            kp.TokenCategoryHierarchyMapper.match(kp.TokenCategory.NOTE, include=None)
        )

    def test_match_with_exclude_only(self):
        """Test that an excluded category correctly returns False."""
        self.assertFalse(
            kp.TokenCategoryHierarchyMapper.match(kp.TokenCategory.DURATION, exclude={kp.TokenCategory.DURATION})
        )

    def test_match_descendant_inclusion(self):
        """Test that a child category correctly matches an ancestor in include set."""
        self.assertTrue(
            kp.TokenCategoryHierarchyMapper.match(kp.TokenCategory.PITCH, include={kp.TokenCategory.NOTE_REST})
        )

    def test_match_descendant_exclusion(self):
        """Test that a descendant category is excluded when its ancestor is in exclude set."""
        self.assertFalse(
            kp.TokenCategoryHierarchyMapper.match(kp.TokenCategory.PITCH, include={kp.TokenCategory.NOTE_REST}, exclude={kp.TokenCategory.NOTE})
        )

    def test_match_category_not_in_include(self):
        """Test that a category is not matched if it is not in the include set or a descendant."""
        self.assertFalse(
            kp.TokenCategoryHierarchyMapper.match(kp.TokenCategory.BARLINES, include={kp.TokenCategory.NOTE_REST})
        )

    def test_match_category_in_exclude_and_include(self):
        """Ensure that exclude takes precedence over include."""
        self.assertFalse(
            kp.TokenCategoryHierarchyMapper.match(kp.TokenCategory.NOTE, include={kp.TokenCategory.CORE}, exclude={kp.TokenCategory.NOTE})
        )

    def test_match_accept_passing_only_one_category(self):
        kp.TokenCategoryHierarchyMapper.match(kp.TokenCategory.NOTE, include=kp.TokenCategory.NOTE_REST)
        kp.TokenCategoryHierarchyMapper.match(kp.TokenCategory.NOTE, include=kp.TokenCategory.CORE)
        kp.TokenCategoryHierarchyMapper.match(kp.TokenCategory.NOTE, exclude=kp.TokenCategory.NOTE_REST)
        kp.TokenCategoryHierarchyMapper.match(kp.TokenCategory.NOTE, exclude=kp.TokenCategory.CORE)

        self.assertTrue(True)  # If no exception is raised, the test passes.

    def test_match_accept_passing_list_or_tuple_of_categories(self):
        kp.TokenCategoryHierarchyMapper.match(kp.TokenCategory.NOTE, include=[kp.TokenCategory.NOTE_REST])
        kp.TokenCategoryHierarchyMapper.match(kp.TokenCategory.NOTE, include=(kp.TokenCategory.NOTE_REST,))
        kp.TokenCategoryHierarchyMapper.match(kp.TokenCategory.NOTE, include=[kp.TokenCategory.NOTE_REST, kp.TokenCategory.CORE])
        kp.TokenCategoryHierarchyMapper.match(kp.TokenCategory.NOTE, include=(kp.TokenCategory.NOTE_REST, kp.TokenCategory.CORE))
        kp.TokenCategoryHierarchyMapper.match(kp.TokenCategory.NOTE, exclude=[kp.TokenCategory.NOTE_REST])
        kp.TokenCategoryHierarchyMapper.match(kp.TokenCategory.NOTE, exclude=(kp.TokenCategory.NOTE_REST,))
        kp.TokenCategoryHierarchyMapper.match(kp.TokenCategory.NOTE, exclude=[kp.TokenCategory.NOTE_REST, kp.TokenCategory.CORE])
        kp.TokenCategoryHierarchyMapper.match(kp.TokenCategory.NOTE, exclude=(kp.TokenCategory.NOTE_REST, kp.TokenCategory.CORE))

        self.assertTrue(True)

    def test_match_raises_error_on_invalid_include(self):
        with self.assertRaises(ValueError):
            kp.TokenCategoryHierarchyMapper.match(kp.TokenCategory.NOTE, include="NOTE_REST")
        with self.assertRaises(ValueError):
            kp.TokenCategoryHierarchyMapper.match(kp.TokenCategory.NOTE, include=[kp.TokenCategory.NOTE_REST, 0])
        with self.assertRaises(ValueError):
            kp.TokenCategoryHierarchyMapper.match(kp.TokenCategory.NOTE, include=[kp.TokenCategory.NOTE_REST, "NOTE"])
        with self.assertRaises(ValueError):
            kp.TokenCategoryHierarchyMapper.match(kp.TokenCategory.NOTE, include=[kp.TokenCategory.NOTE_REST, None])

    def test_hierarchy_mapper_children(self):
        expected_children = {
            kp.TokenCategory.NOTE_REST,
            kp.TokenCategory.CHORD,
            kp.TokenCategory.EMPTY,
            kp.TokenCategory.ERROR
        }
        self.assertSetEqual(expected_children, kp.TokenCategoryHierarchyMapper.children(kp.TokenCategory.CORE))

    def test_hierarchy_mapper_nodes_CORE(self):
        expected_nodes = {
            kp.TokenCategory.NOTE_REST,
            kp.TokenCategory.NOTE,
            kp.TokenCategory.REST,
            kp.TokenCategory.DURATION,
            kp.TokenCategory.CHORD,
            kp.TokenCategory.EMPTY,
            kp.TokenCategory.DECORATION,
            kp.TokenCategory.PITCH,
            kp.TokenCategory.ALTERATION,
            kp.TokenCategory.ERROR,
        }
        self.assertSetEqual(expected_nodes, kp.TokenCategoryHierarchyMapper.nodes(parent=kp.TokenCategory.CORE))

    def test_hierarchy_mapper_nodes_SIGNATURES(self):
        expected_nodes = {
            kp.TokenCategory.CLEF,
            kp.TokenCategory.TIME_SIGNATURE,
            kp.TokenCategory.METER_SYMBOL,
            kp.TokenCategory.KEY_SIGNATURE,
            kp.TokenCategory.KEY_TOKEN,
        }
        self.assertSetEqual(expected_nodes, kp.TokenCategoryHierarchyMapper.nodes(parent=kp.TokenCategory.SIGNATURES))

    def test_hierarchy_mapper_nodes_NOTE_REST(self):
        expected_nodes = {
            kp.TokenCategory.NOTE,
            kp.TokenCategory.REST,
            kp.TokenCategory.DURATION,
            kp.TokenCategory.PITCH,
            kp.TokenCategory.DECORATION,
            kp.TokenCategory.ALTERATION,
        }
        self.assertSetEqual(expected_nodes, kp.TokenCategoryHierarchyMapper.nodes(parent=kp.TokenCategory.NOTE_REST))

    def test_hierarchy_mapper_leaves_CORE(self):
        expected_leaves = {
            kp.TokenCategory.REST,
            kp.TokenCategory.DURATION,
            kp.TokenCategory.CHORD,
            kp.TokenCategory.EMPTY,
            kp.TokenCategory.DECORATION,
            kp.TokenCategory.PITCH,
            kp.TokenCategory.ALTERATION,
            kp.TokenCategory.ERROR,
        }
        self.assertSetEqual(expected_leaves, kp.TokenCategoryHierarchyMapper.leaves(target=kp.TokenCategory.CORE))

    def test_hierarchy_mapper_leaves_SIGNATURES(self):
        expected_leaves = {
            kp.TokenCategory.CLEF,
            kp.TokenCategory.TIME_SIGNATURE,
            kp.TokenCategory.METER_SYMBOL,
            kp.TokenCategory.KEY_SIGNATURE,
            kp.TokenCategory.KEY_TOKEN
        }
        self.assertSetEqual(expected_leaves, kp.TokenCategoryHierarchyMapper.leaves(target=kp.TokenCategory.SIGNATURES))

    def test_hierarchy_mapper_leaves_NOTE_REST(self):
        expected_leaves = {
            kp.TokenCategory.REST,
            kp.TokenCategory.DURATION,
            kp.TokenCategory.PITCH,
            kp.TokenCategory.DECORATION,
            kp.TokenCategory.ALTERATION,
        }
        self.assertSetEqual(expected_leaves, kp.TokenCategoryHierarchyMapper.leaves(target=kp.TokenCategory.NOTE_REST))

    def test_hierarchy_tree(self):
        with open('resource_dir/hierarchy/tree.txt', 'r') as f:
            expected_tree = f.read()

        real_output = kp.TokenCategoryHierarchyMapper.tree()

        self.assertEqual(expected_tree, real_output)

    def test_hierarchy_valid_categories_CORE(self):
        self._is_valid_category(
            include={kp.TokenCategory.CORE},
            exclude=set(),
            expected_categories={
                kp.TokenCategory.CORE,
                kp.TokenCategory.NOTE_REST,
                kp.TokenCategory.NOTE,
                kp.TokenCategory.REST,
                kp.TokenCategory.DURATION,
                kp.TokenCategory.CHORD,
                kp.TokenCategory.EMPTY,
                kp.TokenCategory.DECORATION,
                kp.TokenCategory.PITCH,
                kp.TokenCategory.ALTERATION,
                kp.TokenCategory.ERROR
            }
        )

        # add mocking tests for TokenCategory calls to TokenCategoryHierarchyMapper here
        ...

    def test_hierarchy_valid_categories_SIGNATURES(self):
        self._is_valid_category(
            include={kp.TokenCategory.SIGNATURES},
            exclude=set(),
            expected_categories={
                kp.TokenCategory.SIGNATURES,
                kp.TokenCategory.CLEF,
                kp.TokenCategory.TIME_SIGNATURE,
                kp.TokenCategory.METER_SYMBOL,
                kp.TokenCategory.KEY_SIGNATURE,
                kp.TokenCategory.KEY_TOKEN,
            }
        )

    def test_hierarchy_valid_categories_NOTE_REST(self):
        self._is_valid_category(
            include={kp.TokenCategory.NOTE_REST},
            exclude=set(),
            expected_categories={
                kp.TokenCategory.NOTE_REST,
                kp.TokenCategory.NOTE,
                kp.TokenCategory.REST,
                kp.TokenCategory.DURATION,
                kp.TokenCategory.PITCH,
                kp.TokenCategory.DECORATION,
                kp.TokenCategory.ALTERATION,
            }
        )

    def test_all_TokenCategory_elements_are_in_hierarchy_map(self):
        all_token_category_in_enums = kp.TokenCategory.all()
        all_token_category_in_hierarchy = kp.TokenCategoryHierarchyMapper.all()

        self.assertSetEqual(all_token_category_in_enums, all_token_category_in_hierarchy)

    @patch('kernpy.TokenCategoryHierarchyMapper.tree')
    def test_tree_calls_mapper(self, mock_tree):
        # Setup the mock to return a dummy tree string.
        dummy_tree = "dummy_tree"
        mock_tree.return_value = dummy_tree

        # Call the class method on TokenCategory.
        result = kp.TokenCategory.tree()

        # Assert the mapper’s tree() was called once and with no arguments.
        mock_tree.assert_called_once_with()
        self.assertEqual(result, dummy_tree)

    @patch('kernpy.TokenCategoryHierarchyMapper.is_child')
    def test_is_child_calls_mapper(self, mock_is_child):
        # Setup the mock return value.
        mock_is_child.return_value = True

        # Call is_child on TokenCategory.
        result = kp.TokenCategory.is_child(child=kp.TokenCategory.NOTE,
                                           parent=kp.TokenCategory.CORE)

        # Assert that TokenCategoryHierarchyMapper.is_child was called with the correct kwargs.
        mock_is_child.assert_called_once_with(parent=kp.TokenCategory.CORE,
                                              child=kp.TokenCategory.NOTE)
        self.assertTrue(result)

    @patch('kernpy.TokenCategoryHierarchyMapper.children')
    def test_children_calls_mapper(self, mock_children):
        target = kp.TokenCategory.CORE
        dummy_children = {kp.TokenCategory.NOTE_REST}
        mock_children.return_value = dummy_children

        result = kp.TokenCategory.children(target)
        # Verify that the mapper method was called with the target passed as parent.
        mock_children.assert_called_once_with(parent=target)
        self.assertEqual(result, dummy_children)


    @patch('kernpy.TokenCategoryHierarchyMapper.valid')
    def test_valid_calls_mapper(self, mock_valid):
        include = {kp.TokenCategory.CORE}
        exclude = {kp.TokenCategory.NOTE}
        dummy_valid = {kp.TokenCategory.CHORD}
        mock_valid.return_value = dummy_valid

        result = kp.TokenCategory.valid(include=include, exclude=exclude)
        mock_valid.assert_called_once_with(include=include, exclude=exclude)
        self.assertEqual(result, dummy_valid)

    @patch('kernpy.TokenCategoryHierarchyMapper.leaves')
    def test_leaves_calls_mapper(self, mock_leaves):
        target = kp.TokenCategory.CORE
        dummy_leaves = {kp.TokenCategory.EMPTY}
        mock_leaves.return_value = dummy_leaves

        result = kp.TokenCategory.leaves(target)
        mock_leaves.assert_called_once_with(target=target)
        self.assertEqual(result, dummy_leaves)

    @patch('kernpy.TokenCategoryHierarchyMapper.nodes')
    def test_nodes_calls_mapper(self, mock_nodes):
        target = kp.TokenCategory.CORE
        dummy_nodes = {kp.TokenCategory.NOTE, kp.TokenCategory.CHORD}
        mock_nodes.return_value = dummy_nodes

        result = kp.TokenCategory.nodes(target)
        mock_nodes.assert_called_once_with(parent=target)
        self.assertEqual(result, dummy_nodes)

    @patch('kernpy.TokenCategoryHierarchyMapper.match')
    def test_match_calls_mapper(self, mock_match):
        target = kp.TokenCategory.NOTE
        include = {kp.TokenCategory.CORE}
        exclude = {kp.TokenCategory.NOTE}
        dummy_match = False
        mock_match.return_value = dummy_match

        result = kp.TokenCategory.match(target, include=include, exclude=exclude)
        mock_match.assert_called_once_with(category=target, include=include, exclude=exclude)
        self.assertEqual(result, dummy_match)


class PitchRestTestCase(unittest.TestCase):
    def test_PitchRest_creation_generic(self):
        pitch_rest = kp.PitchRest('c')
        self.assertEqual(pitch_rest.pitch, 'c')
        self.assertEqual(pitch_rest.octave, 4)
        self.assertEqual(pitch_rest.is_rest(), False)

    def test_PitchRest_creation_pitch(self):
        pitch_rest = kp.PitchRest('c')
        self.assertEqual(pitch_rest.pitch, 'c')
        pitch_rest = kp.PitchRest('ccc')
        self.assertEqual(pitch_rest.pitch, 'c')
        pitch_rest = kp.PitchRest('C')
        self.assertEqual(pitch_rest.pitch, 'c')
        pitch_rest = kp.PitchRest('CCCCC')
        self.assertEqual(pitch_rest.pitch, 'c')

    def test_PitchRest_creation_octave(self):
        pitch_rest = kp.PitchRest('c')
        self.assertEqual(pitch_rest.octave, 4)
        pitch_rest = kp.PitchRest('ccc')
        self.assertEqual(pitch_rest.octave, 6)
        pitch_rest = kp.PitchRest('ccccc')
        self.assertEqual(pitch_rest.octave, 8)
        pitch_rest = kp.PitchRest('C')
        self.assertEqual(pitch_rest.octave, 3)
        pitch_rest = kp.PitchRest('CCC')
        self.assertEqual(pitch_rest.octave, 1)
        pitch_rest = kp.PitchRest('CCCCC')
        self.assertEqual(pitch_rest.octave, -1)

    def test_PitchRest_creation_rest(self):
        pitch_rest = kp.PitchRest('r')
        self.assertEqual(pitch_rest.pitch, 'r')
        self.assertEqual(pitch_rest.octave, None)
        self.assertEqual(pitch_rest.is_rest(), True)

    def test_PitchRest_eq(self):
        pitch_rest_a = kp.PitchRest('c')
        pitch_rest_b = kp.PitchRest('c')
        self.assertTrue(pitch_rest_a == pitch_rest_b)

        pitch_rest_a = kp.PitchRest('c')
        pitch_rest_b = kp.PitchRest('ccc')
        self.assertFalse(pitch_rest_a == pitch_rest_b)

        pitch_rest_a = kp.PitchRest('c')
        pitch_rest_b = kp.PitchRest('r')
        self.assertFalse(pitch_rest_a == pitch_rest_b)

        pitch_rest_a = kp.PitchRest('r')
        pitch_rest_b = kp.PitchRest('r')
        self.assertTrue(pitch_rest_a == pitch_rest_b)

    def test_PitchRest_ne(self):
        pitch_rest_a = kp.PitchRest('c')
        pitch_rest_b = kp.PitchRest('d')
        self.assertTrue(pitch_rest_a != pitch_rest_b)

    def test_PitchRest_gt(self):
        pitch_rest_a = kp.PitchRest('c')
        pitch_rest_b = kp.PitchRest('c')
        self.assertFalse(pitch_rest_a > pitch_rest_b)

        pitch_rest_a = kp.PitchRest('c')
        pitch_rest_b = kp.PitchRest('ccc')
        self.assertFalse(pitch_rest_a > pitch_rest_b)

        pitch_rest_a = kp.PitchRest('C')
        pitch_rest_b = kp.PitchRest('C')
        self.assertFalse(pitch_rest_a > pitch_rest_b)

        pitch_rest_a = kp.PitchRest('C')
        pitch_rest_b = kp.PitchRest('CCC')
        self.assertTrue(pitch_rest_a > pitch_rest_b)

        pitch_rest_a = kp.PitchRest('CCC')
        pitch_rest_b = kp.PitchRest('C')
        self.assertFalse(pitch_rest_a > pitch_rest_b)

        pitch_rest_a = kp.PitchRest('r')
        pitch_rest_b = kp.PitchRest('r')
        with self.assertRaises(ValueError):
            pitch_rest_a > pitch_rest_b

    def test_PitchRest_lt(self):
        pitch_rest_a = kp.PitchRest('c')
        pitch_rest_b = kp.PitchRest('c')
        self.assertFalse(pitch_rest_a < pitch_rest_b)

        pitch_rest_a = kp.PitchRest('c')
        pitch_rest_b = kp.PitchRest('ccc')
        self.assertTrue(pitch_rest_a < pitch_rest_b)

        pitch_rest_a = kp.PitchRest('C')
        pitch_rest_b = kp.PitchRest('C')
        self.assertFalse(pitch_rest_a < pitch_rest_b)

        pitch_rest_a = kp.PitchRest('C')
        pitch_rest_b = kp.PitchRest('CCC')
        self.assertFalse(pitch_rest_a < pitch_rest_b)

        pitch_rest_a = kp.PitchRest('CCC')
        pitch_rest_b = kp.PitchRest('C')
        self.assertTrue(pitch_rest_a < pitch_rest_b)

        pitch_rest_a = kp.PitchRest('r')
        pitch_rest_b = kp.PitchRest('r')
        with self.assertRaises(ValueError):
            pitch_rest_a < pitch_rest_b

    def test_PitchRest_ge(self):
        pitch_rest_a = kp.PitchRest('c')
        pitch_rest_b = kp.PitchRest('c')
        self.assertTrue(pitch_rest_a >= pitch_rest_b)

        pitch_rest_a = kp.PitchRest('c')
        pitch_rest_b = kp.PitchRest('ccc')
        self.assertFalse(pitch_rest_a >= pitch_rest_b)

        pitch_rest_a = kp.PitchRest('C')
        pitch_rest_b = kp.PitchRest('C')
        self.assertTrue(pitch_rest_a >= pitch_rest_b)

        pitch_rest_a = kp.PitchRest('C')
        pitch_rest_b = kp.PitchRest('CCC')
        self.assertTrue(pitch_rest_a >= pitch_rest_b)

        pitch_rest_a = kp.PitchRest('CCC')
        pitch_rest_b = kp.PitchRest('C')
        self.assertFalse(pitch_rest_a >= pitch_rest_b)

        pitch_rest_a = kp.PitchRest('r')
        pitch_rest_b = kp.PitchRest('r')
        with self.assertRaises(ValueError):
            pitch_rest_a >= pitch_rest_b

        pitch_rest_a = kp.PitchRest('c')
        pitch_rest_b = kp.PitchRest('r')
        with self.assertRaises(ValueError):
            pitch_rest_a >= pitch_rest_b

    def test_PitchRest_le(self):
        pitch_rest_a = kp.PitchRest('c')
        pitch_rest_b = kp.PitchRest('c')
        self.assertTrue(pitch_rest_a <= pitch_rest_b)

        pitch_rest_a = kp.PitchRest('c')
        pitch_rest_b = kp.PitchRest('ccc')
        self.assertTrue(pitch_rest_a <= pitch_rest_b)

        pitch_rest_a = kp.PitchRest('C')
        pitch_rest_b = kp.PitchRest('C')
        self.assertTrue(pitch_rest_a <= pitch_rest_b)

        pitch_rest_a = kp.PitchRest('C')
        pitch_rest_b = kp.PitchRest('CCC')
        self.assertFalse(pitch_rest_a <= pitch_rest_b)

        pitch_rest_a = kp.PitchRest('CCC')
        pitch_rest_b = kp.PitchRest('C')
        self.assertTrue(pitch_rest_a <= pitch_rest_b)

        pitch_rest_a = kp.PitchRest('r')
        pitch_rest_b = kp.PitchRest('r')
        with self.assertRaises(ValueError):
            pitch_rest_a <= pitch_rest_b

        pitch_rest_a = kp.PitchRest('c')
        pitch_rest_b = kp.PitchRest('r')
        with self.assertRaises(ValueError):
            pitch_rest_a <= pitch_rest_b


class DurationTestCase(unittest.TestCase):
    def test_Duration_creation_generic(self):
        duration = kp.DurationClassical(2)
        self.assertEqual(duration.duration, 2)

        duration = kp.DurationClassical(16)
        self.assertEqual(duration.duration, 16)

        duration = kp.DurationClassical(1)
        self.assertEqual(duration.duration, 1)

        with self.assertRaises(ValueError):
            duration = kp.DurationClassical(0)

        with self.assertRaises(ValueError):
            duration = kp.DurationClassical('abcde')

    def test_Duration_eq(self):
        duration_a = kp.DurationClassical(2)
        duration_b = kp.DurationClassical(2)
        self.assertTrue(duration_a == duration_b)

        duration_a = kp.DurationClassical(2)
        duration_b = kp.DurationClassical(16)
        self.assertFalse(duration_a == duration_b)

    def test_Duration_ne(self):
        duration_a = kp.DurationClassical(2)
        duration_b = kp.DurationClassical(2)
        self.assertFalse(duration_a != duration_b)

        duration_a = kp.DurationClassical(2)
        duration_b = kp.DurationClassical(16)
        self.assertTrue(duration_a != duration_b)

    def test_Duration_gt(self):
        duration_a = kp.DurationClassical(2)
        duration_b = kp.DurationClassical(2)
        self.assertFalse(duration_a > duration_b)

        duration_a = kp.DurationClassical(2)
        duration_b = kp.DurationClassical(16)
        self.assertFalse(duration_a > duration_b)

        duration_a = kp.DurationClassical(16)
        duration_b = kp.DurationClassical(2)
        self.assertTrue(duration_a > duration_b)

    def test_Duration_lt(self):
        duration_a = kp.DurationClassical(2)
        duration_b = kp.DurationClassical(2)
        self.assertFalse(duration_a < duration_b)

        duration_a = kp.DurationClassical(2)
        duration_b = kp.DurationClassical(16)
        self.assertTrue(duration_a < duration_b)

        duration_a = kp.DurationClassical(16)
        duration_b = kp.DurationClassical(2)
        self.assertFalse(duration_a < duration_b)

    def test_Duration_ge(self):
        duration_a = kp.DurationClassical(2)
        duration_b = kp.DurationClassical(2)
        self.assertTrue(duration_a >= duration_b)

        duration_a = kp.DurationClassical(2)
        duration_b = kp.DurationClassical(16)
        self.assertFalse(duration_a >= duration_b)

        duration_a = kp.DurationClassical(16)
        duration_b = kp.DurationClassical(2)
        self.assertTrue(duration_a >= duration_b)

    def test_Duration_le(self):
        duration_a = kp.DurationClassical(2)
        duration_b = kp.DurationClassical(2)
        self.assertTrue(duration_a <= duration_b)

        duration_a = kp.DurationClassical(2)
        duration_b = kp.DurationClassical(16)
        self.assertTrue(duration_a <= duration_b)

        duration_a = kp.DurationClassical(16)
        duration_b = kp.DurationClassical(2)
        self.assertFalse(duration_a <= duration_b)

    def test_Duration_modify_duration(self):
        duration = kp.DurationClassical(2)
        new_duration = duration.modify(4)
        self.assertEqual(new_duration.duration, 8)

        duration = kp.DurationClassical(16)
        new_duration = duration.modify(2)
        self.assertEqual(new_duration.duration, 32)

        duration = kp.DurationClassical(2)
        new_duration = duration.modify(1)
        self.assertEqual(new_duration.duration, 2)

        duration = kp.DurationClassical(2)
        with self.assertRaises(ValueError):
            new_duration = duration.modify(0)

        duration = kp.DurationClassical(2)
        with self.assertRaises(ValueError):
            new_duration = duration.modify(-1)

        duration = kp.DurationClassical(2)
        with self.assertRaises(ValueError):
            new_duration = duration.modify(1.5)


class DurationMensuralTestCase(unittest.TestCase):
    def test_DurationMensural_creation_generic(self):
        pass


class TokenExportTestCase(unittest.TestCase):
    def test_token_export_simple_tokens(self):
        t1 = kp.SimpleToken('a', kp.TokenCategory.CLEF)
        self.assertEqual(t1.export(), 'a')

    def test_token_export_complex_tokens(self):
        t1 = kp.Subtoken('a', kp.TokenCategory.CLEF)
        t2 = kp.Subtoken('b', kp.TokenCategory.TIME_SIGNATURE)
        t3 = kp.Subtoken('c', kp.TokenCategory.INSTRUMENTS)
        compound = kp.CompoundToken(
            encoding='abc',
            category=kp.TokenCategory.STRUCTURAL,
            subtokens=[t1, t2, t3]
        )

        self.assertEqual(compound.export(), 'a@b@c')

    def test_token_export_compund_tokens_with_filter_one(self):
        t1 = kp.Subtoken('a', kp.TokenCategory.PITCH)
        t2 = kp.Subtoken('b', kp.TokenCategory.DURATION)
        t3 = kp.Subtoken('c', kp.TokenCategory.DECORATION)
        compound = kp.CompoundToken(
            encoding='abc',
            category=kp.TokenCategory.STRUCTURAL,
            subtokens=[t1, t2, t3]
        )
        filter_fn = lambda x: x in {kp.TokenCategory.DURATION}
        self.assertEqual('b', compound.export(filter_categories=filter_fn))

    def test_token_export_compund_tokens_with_filter_two(self):
        t1 = kp.Subtoken('a', kp.TokenCategory.PITCH)
        t2 = kp.Subtoken('b', kp.TokenCategory.DURATION)
        t3 = kp.Subtoken('c', kp.TokenCategory.DECORATION)
        compound = kp.CompoundToken(
            encoding='abc',
            category=kp.TokenCategory.STRUCTURAL,
            subtokens=[t1, t2, t3]
        )
        filter_fn = lambda x: x in {kp.TokenCategory.PITCH, kp.TokenCategory.DURATION}
        self.assertEqual('a@b', compound.export(filter_categories=filter_fn))

    def test_token_export_compund_tokens_with_filter_empty(self):
        t1 = kp.Subtoken('a', kp.TokenCategory.PITCH)
        t2 = kp.Subtoken('b', kp.TokenCategory.DURATION)
        t3 = kp.Subtoken('c', kp.TokenCategory.DECORATION)
        compound = kp.CompoundToken(
            encoding='abc',
            category=kp.TokenCategory.STRUCTURAL,
            subtokens=[t1, t2, t3]
        )
        filter_fn = lambda x: x in []
        self.assertEqual('*', compound.export(filter_categories=filter_fn))

    def test_export_note_rest_token_when_lambda_comes_from_function_class(self):
        pitch = kp.Subtoken('c', kp.TokenCategory.PITCH)
        duration = kp.Subtoken('4', kp.TokenCategory.DURATION)

        note_rest = kp.NoteRestToken(
            encoding='c4',
            pitch_duration_subtokens=[pitch, duration],
            decoration_subtokens=[]
        )

        static_ref = lambda x: x in {kp.TokenCategory.PITCH }
        self.assertEqual('c4', note_rest.encoding)
        self.assertEqual('c', note_rest.export(filter_categories=static_ref))

        class TempFilter:
            def __init__(self, categories):
                self.categories = categories
                self.fn_filter = lambda x: x in self.categories

        self.assertEqual('c', note_rest.export(filter_categories=TempFilter({kp.TokenCategory.PITCH}).fn_filter))


    def test_token_export_NoteRestToken(self):
        pitch = kp.Subtoken('c', kp.TokenCategory.PITCH)
        duration = kp.Subtoken('4', kp.TokenCategory.DURATION)
        decoration_1 = kp.Subtoken('!', kp.TokenCategory.DECORATION)
        decoration_2 = kp.Subtoken('?', kp.TokenCategory.DECORATION)

        note_rest = kp.NoteRestToken(
            encoding='c4!?',
            pitch_duration_subtokens=[pitch, duration],
            decoration_subtokens=[decoration_1, decoration_2]
        )
        self.assertEqual('c4!?', note_rest.encoding)
        self.assertEqual('4@c·!·?', note_rest.export())

    def test_token_export_NoteRestToken_without_decorations(self):
        pitch = kp.Subtoken('c', kp.TokenCategory.PITCH)
        duration = kp.Subtoken('4', kp.TokenCategory.DURATION)

        note_rest = kp.NoteRestToken(
            encoding='c4',
            pitch_duration_subtokens=[pitch, duration],
            decoration_subtokens=[]
        )
        self.assertEqual('c4', note_rest.encoding)
        self.assertEqual('4@c', note_rest.export())

    def test_token_export_NoteRestToken_without_pitch_duration_assert_raise(self):
        with self.assertRaises(ValueError):
            note_rest = kp.NoteRestToken(
                encoding='c4',
                pitch_duration_subtokens=[],
                decoration_subtokens=[]
            )

    def test_token_export_NoteRestToken_without_decoration_assert_not_raise(self):
        pitch = kp.Subtoken('c', kp.TokenCategory.PITCH)
        duration = kp.Subtoken('4', kp.TokenCategory.DURATION)

        note_rest = kp.NoteRestToken(
            encoding='c4',
            pitch_duration_subtokens=[pitch, duration],
            decoration_subtokens=[]
        )
        self.assertEqual('c4', note_rest.encoding)
        self.assertEqual('4@c', note_rest.export())

    def test_token_export_NoteRestToken_encoding_wrong_order(self):
        pitch = kp.Subtoken('c', kp.TokenCategory.PITCH)
        duration = kp.Subtoken('4', kp.TokenCategory.DURATION)
        decoration_1 = kp.Subtoken('!', kp.TokenCategory.DECORATION)
        decoration_2 = kp.Subtoken('?', kp.TokenCategory.DECORATION)

        note_rest_1 = kp.NoteRestToken(
            encoding='4c!?',
            pitch_duration_subtokens=[pitch, duration],
            decoration_subtokens=[decoration_1, decoration_2]
        )
        note_rest_1_2 = kp.NoteRestToken(
            encoding='4c!?',
            pitch_duration_subtokens=[duration, pitch],
            decoration_subtokens=[decoration_1, decoration_2]
        )
        note_rest_2 = kp.NoteRestToken(
            encoding='c!?4',
            pitch_duration_subtokens=[pitch, duration],
            decoration_subtokens=[decoration_1, decoration_2]
        )
        note_rest_3 = kp.NoteRestToken(
            encoding='!?4c',
            pitch_duration_subtokens=[pitch, duration],
            decoration_subtokens=[decoration_1, decoration_2]
        )
        note_rest_4 = kp.NoteRestToken(
            encoding='4c!?',
            pitch_duration_subtokens=[pitch, duration],
            decoration_subtokens=[decoration_1, decoration_2]
        )
        note_rest_5 = kp.NoteRestToken(
            encoding='c4!?',
            pitch_duration_subtokens=[pitch, duration],
            decoration_subtokens=[decoration_1, decoration_2]
        )


        self.assertEqual('4c!?', note_rest_1.encoding)
        self.assertEqual('4@c·!·?', note_rest_1.export())
        self.assertEqual('4c!?', note_rest_1_2.encoding)
        self.assertEqual('4@c·!·?', note_rest_1_2.export())
        self.assertEqual('c!?4', note_rest_2.encoding)
        self.assertEqual('4@c·!·?', note_rest_2.export())
        self.assertEqual('!?4c', note_rest_3.encoding)
        self.assertEqual('4@c·!·?', note_rest_3.export())
        self.assertEqual('4c!?', note_rest_4.encoding)
        self.assertEqual('4@c·!·?', note_rest_4.export())
        self.assertEqual('c4!?', note_rest_5.encoding)
        self.assertEqual('4@c·!·?', note_rest_5.export())

